#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This operation implements screen plotting

from losoto.lib_operations import *
from losoto.operations.directionscreen import _calc_piercepoint
from losoto._logging import logger as logging

logging.debug('Loading PLOTSCREEN module.')

def _run_parser(soltab, parser, step):
    resSoltab = parser.getstr( step, "resSoltab", '' )
    minZ, maxZ = parser.getarrayfloat( step, "MinMax", [0.0, 0.0] )
    prefix = parser.getstr( step, "Prefix", '' )
    remove_gradient = parser.getbool( step, "RemoveGradient", False )
    show_source_names = parser.getbool( step, "ShowSourceNames", False )
    ncpu = parser.getint( step, "npcu", 0 )

    parser.checkSpelling( step, soltab, ['resSoltab', 'minZ', 'maxZ', 'prefix', 'remove_gradient', 'show_source_names'])
    return run(soltab, resSoltab, minZ, maxZ, prefix, remove_gradient, show_source_names, ncpu)


def _phase_cm():
    """
    Returns colormap for phase plots
    """
    from matplotlib.colors import ListedColormap

    cm_data = [[ 0.65830839, 0.46993917, 0.04941288],
               [ 0.66433742, 0.4662019 , 0.05766473],
               [ 0.67020869, 0.46248014, 0.0653456 ],
               [ 0.67604299, 0.45869838, 0.07273174],
               [ 0.68175228, 0.45491407, 0.07979262],
               [ 0.6874028 , 0.45108417, 0.08667103],
               [ 0.6929505 , 0.44723893, 0.09335869],
               [ 0.69842619, 0.44335768, 0.09992839],
               [ 0.7038123 , 0.43945328, 0.1063871 ],
               [ 0.70912069, 0.43551765, 0.11277174],
               [ 0.71434524, 0.43155576, 0.11909348],
               [ 0.71949289, 0.42756272, 0.12537606],
               [ 0.72455619, 0.4235447 , 0.13162325],
               [ 0.72954895, 0.41949098, 0.13786305],
               [ 0.73445172, 0.41541774, 0.14408039],
               [ 0.73929496, 0.41129973, 0.15032217],
               [ 0.74403834, 0.40717158, 0.15654335],
               [ 0.74873695, 0.40298519, 0.16282282],
               [ 0.75332319, 0.39880107, 0.16907566],
               [ 0.75788083, 0.39454245, 0.17542179],
               [ 0.7623326 , 0.39028096, 0.18175915],
               [ 0.76673205, 0.38596549, 0.18816819],
               [ 0.77105247, 0.38162141, 0.19461532],
               [ 0.77529528, 0.37724732, 0.20110652],
               [ 0.77948666, 0.37281509, 0.2076873 ],
               [ 0.78358534, 0.36836772, 0.21429736],
               [ 0.78763763, 0.363854  , 0.22101648],
               [ 0.79161134, 0.35930804, 0.2277974 ],
               [ 0.79550606, 0.3547299 , 0.23464353],
               [ 0.79935398, 0.35007959, 0.24161832],
               [ 0.80311671, 0.34540152, 0.24865892],
               [ 0.80681033, 0.34067452, 0.25580075],
               [ 0.8104452 , 0.33588248, 0.26307222],
               [ 0.8139968 , 0.33105538, 0.27043183],
               [ 0.81747689, 0.32617526, 0.27791096],
               [ 0.82089415, 0.32122629, 0.28553846],
               [ 0.82422713, 0.3162362 , 0.29327617],
               [ 0.82747661, 0.31120154, 0.30113388],
               [ 0.83066399, 0.30608459, 0.30917579],
               [ 0.83376307, 0.30092244, 0.31734921],
               [ 0.83677286, 0.29571346, 0.32566199],
               [ 0.83969693, 0.29044723, 0.33413665],
               [ 0.84253873, 0.28511151, 0.34279962],
               [ 0.84528297, 0.27972917, 0.35162078],
               [ 0.84792704, 0.27430045, 0.36060681],
               [ 0.85046793, 0.26882624, 0.36976395],
               [ 0.85291056, 0.26328859, 0.37913116],
               [ 0.855242  , 0.25770888, 0.38868217],
               [ 0.85745673, 0.25209367, 0.39841601],
               [ 0.85955023, 0.24644737, 0.40833625],
               [ 0.86151767, 0.24077563, 0.41844557],
               [ 0.86335392, 0.23508521, 0.42874606],
               [ 0.86505685, 0.22937288, 0.43926008],
               [ 0.86661606, 0.22366308, 0.44996127],
               [ 0.86802578, 0.21796785, 0.46084758],
               [ 0.86928003, 0.21230132, 0.47191554],
               [ 0.87037274, 0.20667988, 0.48316015],
               [ 0.87129781, 0.2011224 , 0.49457479],
               [ 0.87204914, 0.19565041, 0.50615118],
               [ 0.87262076, 0.19028829, 0.51787932],
               [ 0.87300686, 0.18506334, 0.5297475 ],
               [ 0.8732019 , 0.18000588, 0.54174232],
               [ 0.87320066, 0.1751492 , 0.55384874],
               [ 0.87299833, 0.17052942, 0.56605016],
               [ 0.87259058, 0.16618514, 0.57832856],
               [ 0.87197361, 0.16215698, 0.59066466],
               [ 0.87114414, 0.15848667, 0.60303881],
               [ 0.87009966, 0.15521687, 0.61542844],
               [ 0.86883823, 0.15238892, 0.62781175],
               [ 0.86735858, 0.15004199, 0.64016651],
               [ 0.8656601 , 0.14821149, 0.65247022],
               [ 0.86374282, 0.14692762, 0.66470043],
               [ 0.86160744, 0.14621386, 0.67683495],
               [ 0.85925523, 0.14608582, 0.68885204],
               [ 0.85668805, 0.14655046, 0.70073065],
               [ 0.85390829, 0.14760576, 0.71245054],
               [ 0.85091881, 0.14924094, 0.7239925 ],
               [ 0.84772287, 0.15143717, 0.73533849],
               [ 0.84432409, 0.15416865, 0.74647174],
               [ 0.84072639, 0.15740403, 0.75737678],
               [ 0.83693394, 0.16110786, 0.76803952],
               [ 0.83295108, 0.16524205, 0.77844723],
               [ 0.82878232, 0.16976729, 0.78858858],
               [ 0.82443225, 0.17464414, 0.7984536 ],
               [ 0.81990551, 0.179834  , 0.80803365],
               [ 0.81520674, 0.18529984, 0.8173214 ],
               [ 0.81034059, 0.19100664, 0.82631073],
               [ 0.80531176, 0.1969216 , 0.83499645],
               [ 0.80012467, 0.20301465, 0.84337486],
               [ 0.79478367, 0.20925826, 0.8514432 ],
               [ 0.78929302, 0.21562737, 0.85919957],
               [ 0.78365681, 0.22209936, 0.86664294],
               [ 0.77787898, 0.22865386, 0.87377308],
               [ 0.7719633 , 0.23527265, 0.88059043],
               [ 0.76591335, 0.24193947, 0.88709606],
               [ 0.7597325 , 0.24863985, 0.89329158],
               [ 0.75342394, 0.25536094, 0.89917908],
               [ 0.74699063, 0.26209137, 0.90476105],
               [ 0.74043533, 0.2688211 , 0.91004033],
               [ 0.73376055, 0.27554128, 0.91502   ],
               [ 0.72696862, 0.28224415, 0.91970339],
               [ 0.7200616 , 0.2889229 , 0.92409395],
               [ 0.71304134, 0.29557159, 0.92819525],
               [ 0.70590945, 0.30218508, 0.9320109 ],
               [ 0.69866732, 0.30875887, 0.93554451],
               [ 0.69131609, 0.31528914, 0.93879964],
               [ 0.68385669, 0.32177259, 0.94177976],
               [ 0.6762898 , 0.32820641, 0.94448822],
               [ 0.6686159 , 0.33458824, 0.94692818],
               [ 0.66083524, 0.3409161 , 0.94910264],
               [ 0.65294785, 0.34718834, 0.95101432],
               [ 0.64495358, 0.35340362, 0.95266571],
               [ 0.63685208, 0.35956083, 0.954059  ],
               [ 0.62864284, 0.3656591 , 0.95519608],
               [ 0.62032517, 0.3716977 , 0.95607853],
               [ 0.61189825, 0.37767607, 0.95670757],
               [ 0.60336117, 0.38359374, 0.95708408],
               [ 0.59471291, 0.3894503 , 0.95720861],
               [ 0.58595242, 0.39524541, 0.95708134],
               [ 0.5770786 , 0.40097871, 0.95670212],
               [ 0.56809041, 0.40664983, 0.95607045],
               [ 0.55898686, 0.41225834, 0.95518556],
               [ 0.54976709, 0.41780374, 0.95404636],
               [ 0.5404304 , 0.42328541, 0.95265153],
               [ 0.53097635, 0.42870263, 0.95099953],
               [ 0.52140479, 0.43405447, 0.94908866],
               [ 0.51171597, 0.43933988, 0.94691713],
               [ 0.50191056, 0.44455757, 0.94448311],
               [ 0.49198981, 0.44970607, 0.94178481],
               [ 0.48195555, 0.45478367, 0.93882055],
               [ 0.47181035, 0.45978843, 0.93558888],
               [ 0.46155756, 0.46471821, 0.93208866],
               [ 0.45119801, 0.46957218, 0.92831786],
               [ 0.44073852, 0.47434688, 0.92427669],
               [ 0.43018722, 0.47903864, 0.9199662 ],
               [ 0.41955166, 0.4836444 , 0.91538759],
               [ 0.40884063, 0.48816094, 0.91054293],
               [ 0.39806421, 0.49258494, 0.90543523],
               [ 0.38723377, 0.49691301, 0.90006852],
               [ 0.37636206, 0.50114173, 0.89444794],
               [ 0.36546127, 0.5052684 , 0.88857877],
               [ 0.35454654, 0.5092898 , 0.88246819],
               [ 0.34363779, 0.51320158, 0.87612664],
               [ 0.33275309, 0.51700082, 0.86956409],
               [ 0.32191166, 0.52068487, 0.86279166],
               [ 0.31113372, 0.52425144, 0.85582152],
               [ 0.3004404 , 0.52769862, 0.84866679],
               [ 0.28985326, 0.53102505, 0.84134123],
               [ 0.27939616, 0.53422931, 0.83386051],
               [ 0.26909181, 0.53731099, 0.82623984],
               [ 0.258963  , 0.5402702 , 0.81849475],
               [ 0.24903239, 0.54310763, 0.8106409 ],
               [ 0.23932229, 0.54582448, 0.80269392],
               [ 0.22985664, 0.54842189, 0.79467122],
               [ 0.2206551 , 0.55090241, 0.78658706],
               [ 0.21173641, 0.55326901, 0.77845533],
               [ 0.20311843, 0.55552489, 0.77028973],
               [ 0.1948172 , 0.55767365, 0.76210318],
               [ 0.1868466 , 0.55971922, 0.75390763],
               [ 0.17921799, 0.56166586, 0.74571407],
               [ 0.1719422 , 0.56351747, 0.73753498],
               [ 0.16502295, 0.56527915, 0.72937754],
               [ 0.15846116, 0.566956  , 0.72124819],
               [ 0.15225499, 0.56855297, 0.71315321],
               [ 0.14639876, 0.57007506, 0.70509769],
               [ 0.14088284, 0.57152729, 0.69708554],
               [ 0.13569366, 0.57291467, 0.68911948],
               [ 0.13081385, 0.57424211, 0.68120108],
               [ 0.12622247, 0.57551447, 0.67333078],
               [ 0.12189539, 0.57673644, 0.66550792],
               [ 0.11780654, 0.57791235, 0.65773233],
               [ 0.11392613, 0.5790468 , 0.64999984],
               [ 0.11022348, 0.58014398, 0.64230637],
               [ 0.10666732, 0.58120782, 0.63464733],
               [ 0.10322631, 0.58224198, 0.62701729],
               [ 0.0998697 , 0.58324982, 0.61941001],
               [ 0.09656813, 0.58423445, 0.61181853],
               [ 0.09329429, 0.58519864, 0.60423523],
               [ 0.09002364, 0.58614483, 0.5966519 ],
               [ 0.08673514, 0.58707512, 0.58905979],
               [ 0.08341199, 0.58799127, 0.58144971],
               [ 0.08004245, 0.58889466, 0.57381211],
               [ 0.07662083, 0.58978633, 0.56613714],
               [ 0.07314852, 0.59066692, 0.55841474],
               [ 0.06963541, 0.5915367 , 0.55063471],
               [ 0.06610144, 0.59239556, 0.54278681],
               [ 0.06257861, 0.59324304, 0.53486082],
               [ 0.05911304, 0.59407833, 0.52684614],
               [ 0.05576765, 0.5949003 , 0.5187322 ],
               [ 0.05262511, 0.59570732, 0.51050978],
               [ 0.04978881, 0.5964975 , 0.50216936],
               [ 0.04738319, 0.59726862, 0.49370174],
               [ 0.04555067, 0.59801813, 0.48509809],
               [ 0.04444396, 0.59874316, 0.47635   ],
               [ 0.04421323, 0.59944056, 0.46744951],
               [ 0.04498918, 0.60010687, 0.45838913],
               [ 0.04686604, 0.60073837, 0.44916187],
               [ 0.04988979, 0.60133103, 0.43976125],
               [ 0.05405573, 0.60188055, 0.4301812 ],
               [ 0.05932209, 0.60238289, 0.42040543],
               [ 0.06560774, 0.60283258, 0.41043772],
               [ 0.07281962, 0.60322442, 0.40027363],
               [ 0.08086177, 0.60355283, 0.38990941],
               [ 0.08964366, 0.60381194, 0.37934208],
               [ 0.09908952, 0.60399554, 0.36856412],
               [ 0.10914617, 0.60409695, 0.35755799],
               [ 0.11974119, 0.60410858, 0.34634096],
               [ 0.13082746, 0.6040228 , 0.33491416],
               [ 0.14238003, 0.60383119, 0.323267  ],
               [ 0.1543847 , 0.60352425, 0.31138823],
               [ 0.16679093, 0.60309301, 0.29931029],
               [ 0.17959757, 0.60252668, 0.2870237 ],
               [ 0.19279966, 0.60181364, 0.27452964],
               [ 0.20634465, 0.60094466, 0.2618794 ],
               [ 0.22027287, 0.5999043 , 0.24904251],
               [ 0.23449833, 0.59868591, 0.23611022],
               [ 0.24904416, 0.5972746 , 0.2230778 ],
               [ 0.26382006, 0.59566656, 0.21004673],
               [ 0.2788104 , 0.5938521 , 0.19705484],
               [ 0.29391494, 0.59183348, 0.18421621],
               [ 0.3090634 , 0.58961302, 0.17161942],
               [ 0.32415577, 0.58720132, 0.15937753],
               [ 0.3391059 , 0.58461164, 0.14759012],
               [ 0.35379624, 0.58186793, 0.13637734],
               [ 0.36817905, 0.5789861 , 0.12580054],
               [ 0.38215966, 0.57599512, 0.1159504 ],
               [ 0.39572824, 0.57290928, 0.10685038],
               [ 0.40881926, 0.56975727, 0.09855521],
               [ 0.42148106, 0.56654159, 0.09104002],
               [ 0.43364953, 0.56329296, 0.08434116],
               [ 0.44538908, 0.56000859, 0.07841305],
               [ 0.45672421, 0.5566943 , 0.07322913],
               [ 0.46765017, 0.55336373, 0.06876762],
               [ 0.47819138, 0.5500213 , 0.06498436],
               [ 0.48839686, 0.54666195, 0.06182163],
               [ 0.49828924, 0.5432874 , 0.05922726],
               [ 0.50789114, 0.53989827, 0.05714466],
               [ 0.51722475, 0.53649429, 0.05551476],
               [ 0.5263115 , 0.53307443, 0.05427793],
               [ 0.53517186, 0.52963707, 0.05337567],
               [ 0.54382515, 0.52618009, 0.05275208],
               [ 0.55228947, 0.52270103, 0.05235479],
               [ 0.56058163, 0.51919713, 0.0521356 ],
               [ 0.56871719, 0.51566545, 0.05205062],
               [ 0.57671045, 0.51210292, 0.0520602 ],
               [ 0.5845745 , 0.50850636, 0.05212851],
               [ 0.59232129, 0.50487256, 0.05222299],
               [ 0.5999617 , 0.50119827, 0.05231367],
               [ 0.60750568, 0.49748022, 0.05237234],
               [ 0.61496232, 0.49371512, 0.05237168],
               [ 0.62233999, 0.48989963, 0.05228423],
               [ 0.62964652, 0.48603032, 0.05208127],
               [ 0.63688935, 0.48210362, 0.05173155],
               [ 0.64407572, 0.4781157 , 0.0511996 ],
               [ 0.65121289, 0.47406244, 0.05044367],
               [ 0.65830839, 0.46993917, 0.04941288]]
    cm = ListedColormap(cm_data, name=__file__)

    return cm


def _calculate_screen(inscreen, residuals, pp, N_piercepoints, k, east, north, up,
    T, Nx, Ny, sindx, height, beta_val, r_0, is_phase, outQueue):
    """
    Calculates screen images

    Parameters
    ----------
    inscreen : array
        Array of screen values at the piercepoints
    residuals : array
        Array of screen residuals at the piercepoints
    pp : array
        Array of piercepoint locations
    N_piercepoints : int
        Number of pierce points
    k : int
        Time index
    east : array
        East array
    north : array
        North array
    up : array
        Up array
    T : array
        T array
    Nx : int
        Number of pixels in x for screen
    Ny : int
        Number of pixels in y for screen
    sindx : int
        Station index
    height : float
        height of screen (m)
    beta_val : float
        power-law index for phase structure function (5/3 =>
        pure Kolmogorov turbulence)
    r_0 : float
        scale size of phase fluctuations
    is_phase : bool
        input screen is a phase screen

    """
    from numpy import kron, concatenate, newaxis
    from numpy.linalg import pinv, norm
    import numpy as np

    screen = np.zeros((Nx, Ny))

    if height == 0.0:
        pp1 = pp[:, :]
    else:
        pp1 = np.dot(pp[:, :], T)

    min_xy = np.amin(pp1, axis=0)
    max_xy = np.amax(pp1, axis=0)
    extent = max_xy - min_xy
    lowerk = min_xy - 0.1 * extent
    upperk = max_xy + 0.1 * extent
    im_extent_mk = upperk - lowerk
    pix_per_mk = Nx / im_extent_mk[0]
    m_per_pixk = 1.0 / pix_per_mk

    xr = np.arange(lowerk[0], upperk[0], m_per_pixk)
    yr = np.arange(lowerk[1], upperk[1], m_per_pixk)
    D = np.resize(pp, (N_piercepoints, N_piercepoints, 3))
    D = np.transpose(D, (1, 0, 2)) - D
    D2 = np.sum(D**2, axis=2)
    C = -(D2 / r_0**2)**(beta_val / 2.0) / 2.0
    f = inscreen.reshape(N_piercepoints)
    fitted_tec = np.dot(C, f) + residuals
    if is_phase:
        fitted_tec = normalize_phase(fitted_tec)
    for i, xi in enumerate(xr[0: Nx]):
        for j, yi in enumerate(yr[0: Ny]):
            if height == 0.0:
                p = np.array([xi, yi, 0.0])
            else:
                p, airmass = _calc_piercepoint(np.dot(np.array([xi, yi]), np.array([east, north])), up, height)
            d2 = np.sum(np.square(pp - p), axis=1)
            c = -(d2 / ( r_0**2 ))**(beta_val / 2.0) / 2.0
            screen[i, j] = np.dot(c, f)

    # Calculate the piercepoint coords, normalized to 1
    if height == 0.0:
        x = pp1[:, 0]
        y = pp1[:, 1]
    else:
        x = (pp1[:, 0] / 1000.0 - min_xy[0] / 1000.0) / extent[0] # km
        y = (pp1[:, 1] / 1000.0 - min_xy[1] / 1000.0) / extent[1] # km

    outQueue.put([k, fitted_tec, screen, x, y])


def _plot_frame(screen, fitted_phase1, residuals, weights, x, y, k, lower,
    upper, vmin, vmax, source_names, show_source_names, station_names, sindx,
    root_dir, prestr, is_image_plane,  midRA, midDec, order, is_phase, outQueue):
    """
    Plots screen images

    Parameters
    ----------
    screen : array
        Image of screen values
    fitted_phase1 : array
        Array of fitted phase values
    residuals : array
        Array of phase residuals at the piercepoints
    weights : array
        Array of weights at the piercepoints
    x : array
        Array of piercepoint x locations
    y : array
        Array of piercepoint y locations
    k : int
        Time index
    lower : array
        Array of lower limits for plot
    upper : array
        Array of upper limits for plot
    vmin : float
        minimum value for plot range
    vmax : float
        maximum value for plot range
    source_names : list
        List of source (direction) names
    show_source_names : bool
        label sources on screen plots
    order : int
        order of screen
    is_phase : bool
        True if screens are phase screens

    """
    if not 'matplotlib' in sys.modules:
        import matplotlib as mpl
        mpl.rc('figure.subplot',left=0.05, bottom=0.05, right=0.95, top=0.95,wspace=0.22, hspace=0.22 )
        mpl.use("Agg")
    import matplotlib as mpl
    import matplotlib.pyplot as plt # after setting "Agg" to speed up
    from losoto.operations.stationscreen import _makeWCS, _circ_chi2
    import numpy as np
    try:
        try:
            from astropy.visualization.wcsaxes import WCSAxes
            hasWCSaxes = True
        except:
            from wcsaxes import WCSAxes
            hasWCSaxes = True
    except:
        hasWCSaxes = False
    from matplotlib.colors import LinearSegmentedColormap


    fig = plt.figure(figsize=(6,6))

    # Set colormap
    if is_phase:
        cmap = _phase_cm()
    else:
        cmap = plt.cm.jet
    sm = plt.cm.ScalarMappable(cmap=cmap,
        norm=mpl.colors.Normalize(vmin=vmin, vmax=vmax))
    sm._A = []

    if is_image_plane and hasWCSaxes:
        wcs = _makeWCS(midRA, midDec)
        ax = fig.add_axes([0.15, 0.1, 0.8, 0.8], projection=wcs)
    else:
        plt.gca().set_aspect('equal')
        ax = plt.gca()

    s = []
    c = []
    xf = []
    yf = []
    weights = np.array(weights, dtype=float)
    nonflagged = np.where(weights > 0.0)
    for j in range(fitted_phase1.shape[0]):
        if weights[j] > 0.0:
            s.append(max(20, 200*np.sqrt(weights[j]/np.median(weights[nonflagged]))))
        else:
            s.append(120)
            xf.append(x[j])
            yf.append(y[j])
        c.append(sm.to_rgba(fitted_phase1[j]))

    if is_image_plane:
        min_x = np.min(x)
        max_x = np.max(x)
        min_y = np.min(y)
        max_y = np.max(y)
        extent_x = max_x - min_x
        extent_y = max_y - min_y
        lower = [min_x - 0.1 * extent_x, min_y - 0.1 * extent_y]
        upper = [max_x + 0.1 * extent_x, max_y + 0.1 * extent_y]
        Nx = screen.shape[0]
        pix_per_m = Nx / (upper[0] - lower[0])
        m_per_pix = 1.0 / pix_per_m
        xr = np.arange(lower[0], upper[0], m_per_pix)
        yr = np.arange(lower[1], upper[1], m_per_pix)
        lower = np.array([xr[0], yr[0]])
        upper = np.array([xr[-1], yr[-1]])
    else:
        # convert from m to km
        lower /= 1000.0
        upper /= 1000.0

    im = ax.imshow(screen.transpose([1, 0])[:, :],
        cmap = cmap,
        origin = 'lower',
        interpolation = 'nearest',
        extent = (lower[0], upper[0], lower[1], upper[1]),
        vmin=vmin, vmax=vmax)

    cbar = plt.colorbar(im)
    cbar.set_label('Value', rotation=270)

    ax.scatter(np.array(x), np.array(y), s=np.array(s), c=np.array(c), alpha=0.7, cmap=cmap, vmin=vmin, vmax=vmax, edgecolor='black')
    if len(xf) > 0:
        ax.scatter(xf, yf, s=120, c='k', marker='x')
    if show_source_names:
        labels = source_names
        for label, xl, yl in zip(labels, x, y):
            plt.annotate(
                label,
                xy = (xl, yl), xytext = (-2, 2),
                textcoords = 'offset points', ha = 'right', va = 'bottom')

    nsrcs = np.where(weights > 0.0)[0].size
    if is_phase:
        redchi2 =  _circ_chi2(residuals, weights) / (nsrcs-order)
    else:
        redchi2 =  np.sum(np.square(residuals) * weights) / (nsrcs-order)
    if sindx >= 0:
        plt.title('Station {0}, Time {1} (red. chi2 = {2:0.3f})'.format(station_names[sindx], k, redchi2))
    else:
        plt.title('Time {0}'.format(k))
    if is_image_plane:
        ax.set_xlim(lower[0], upper[0])
        ax.set_ylim(lower[1], upper[1])
        ax.set_aspect('equal')
        if hasWCSaxes:
            RAAxis = ax.coords['ra']
            RAAxis.set_axislabel('RA', minpad=0.75)
            RAAxis.set_major_formatter('hh:mm:ss')
            DecAxis = ax.coords['dec']
            DecAxis.set_axislabel('Dec', minpad=0.75)
            DecAxis.set_major_formatter('dd:mm:ss')
            ax.coords.grid(color='black', alpha=0.5, linestyle='solid')
            plt.xlabel("RA")
            plt.ylabel("Dec")
        else:
            plt.xlabel("RA (arb. units)")
            plt.ylabel("Dec (arb. units)")
    else:
        # Reverse the axis so that RA coord increases to left
        plt.xlim(upper[0], lower[0])
        plt.ylim(lower[1], upper[1])
        plt.xlabel('Projected Distance East-West (km)')
        plt.ylabel('Projected Distance North-South (km)')
    if sindx >= 0:
        plt.savefig(root_dir + '/' + prestr + '_station%0.4i' % sindx + '_frame%0.4i.png' % k, bbox_inches='tight')
    else:
        plt.savefig(root_dir + '/' + prestr + '_frame%0.4i.png' % k, bbox_inches='tight')
    plt.close(fig)


def _make_screen_plots(pp, inscreen, inresiduals, weights, station_names,
    station_positions, source_names, times, height, station_order, beta_val,
    r_0, prefix='frame_', remove_gradient=True, show_source_names=False,
    min_val=None, max_val=None, is_phase=False, midRA=0.0, midDec=0.0, ncpu=0):
    """
    Makes plots of screens

    Parameters
    ----------
    pp : array
        Array of piercepoint locations
    inscreen : array
        Array of screen values at the piercepoints with order [dir, time, ant]
    residuals : array
        Array of screen residuals at the piercepoints with order [dir, time, ant]
    weights : array
        Array of weights for each piercepoint with order [dir, time, ant]
    source_names: array
        Array of source names
    times : array
        Array of times
    height : float
        Height of screen (m)
    station_order : list
        List of order of screens per station (e.g., number of KL base vectors to keep)
    r_0 : float
        Scale size of phase fluctuations
    beta_val : float
        Power-law index for phase structure function
    prefix : str
        Prefix for output file names
    remove_gradient : bool
        Fit and remove a gradient from each screen
    show_source_names : bool
        Label sources on screen plots
    min_val : float
        Minimum value for plot range
    max_val : float
        Maximum value for plot range
    is_phase : bool
        Input screen is a phase screen
    midRA : float
        RA for WCS reference in degrees
    midDec : float
        Dec for WCS reference in degrees
    ncpu : int
        Number of CPUs to use

    """
    from numpy import kron, concatenate, newaxis
    from numpy.linalg import pinv, norm
    import numpy as np
    import os

    # avoids error if re-setting "agg" a second run of plot
    if not 'matplotlib' in sys.modules:
        import matplotlib as mpl
        mpl.rc('figure.subplot',left=0.05, bottom=0.05, right=0.95, top=0.95,wspace=0.22, hspace=0.22 )
        mpl.use("Agg")
    import matplotlib as mpl
    import matplotlib.pyplot as plt # after setting "Agg" to speed up
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes

    # input check
    root_dir = os.path.dirname(prefix)
    if root_dir == '':
        root_dir = './'
    prestr = os.path.basename(prefix) + 'screen'
    try:
        os.makedirs(root_dir)
    except OSError:
        pass

    if height == 0.0:
        N_stations = 1 # screens are single-station screens
    else:
        N_stations = len(station_names)
    N_sources = len(source_names)
    N_times = len(times)
    N_piercepoints = N_sources * N_stations
    xp, yp, zp = station_positions[0, :] # use first station
    east = np.array([-yp, xp, 0])
    east = east / norm(east)

    north = np.array([-xp, -yp, (xp*xp + yp*yp)/zp])
    north = north / norm(north)

    up = np.array([xp, yp, zp])
    up = up / norm(up)

    T = concatenate([east[:, newaxis], north[:, newaxis]], axis=1)

    # Use pierce point locations of first and last time slots to estimate
    # required size of plot in meters
    if height == 0.0:
        is_image_plane = True # pierce points are image plane coords
        pp1_0 = pp[:, 0:2]
        pp1_1 = pp[:, 0:2]
    else:
        is_image_plane = False
        pp1_0 = np.dot(pp[0, :, :], T)
        pp1_1 = np.dot(pp[-1, :, :], T)

    max_xy = np.amax(pp1_0, axis=0) - np.amin(pp1_0, axis=0)
    max_xy_1 = np.amax(pp1_1, axis=0) - np.amin(pp1_1, axis=0)
    if max_xy_1[0] > max_xy[0]:
        max_xy[0] = max_xy_1[0]
    if max_xy_1[1] > max_xy[1]:
        max_xy[1] = max_xy_1[1]

    min_xy = np.array([0.0, 0.0])
    extent = max_xy - min_xy
    lower = min_xy - 0.1 * extent
    upper = max_xy + 0.1 * extent
    im_extent_m = upper - lower
    fitted_phase1 = np.zeros((N_piercepoints, N_times))

    Nx = 60 # set approximate number of pixels in screen
    pix_per_m = Nx / im_extent_m[0]
    m_per_pix = 1.0 / pix_per_m
    xr = np.arange(lower[0], upper[0], m_per_pix)
    yr = np.arange(lower[1], upper[1], m_per_pix)
    Nx = len(xr)
    Ny = len(yr)
    lower = np.array([xr[0], yr[0]])
    upper = np.array([xr[-1], yr[-1]])

    x = np.zeros((N_times, N_piercepoints)) # plot x pos of piercepoints
    y = np.zeros((N_times, N_piercepoints)) # plot y pos of piercepoints
    screen = np.zeros((Nx, Ny, N_times))

    if height == 0.0:
        for sindx in range(station_positions.shape[0]):
            logging.info('Calculating screen images...')
            residuals = inresiduals[:, :, sindx, newaxis].transpose([0, 2, 1]).reshape(N_piercepoints, N_times)
            mpm = multiprocManager(ncpu, _calculate_screen)
            for k in range(N_times):
                mpm.put([inscreen[:, k, sindx], residuals[:, k], pp,
                    N_piercepoints, k, east, north, up, T, Nx, Ny, sindx, height,
                    beta_val, r_0, is_phase])
            mpm.wait()
            for (k, ft, scr, xa, ya) in mpm.get():
                screen[:, :, k] = scr
                fitted_phase1[:, k] = ft
                if is_image_plane:
                    x[k, :] = xa
                    y[k, :] = ya
                else:
                    x[k, :] = xa - np.amin(xa) # remove offsets for each time slot
                    y[k, :] = ya - np.amin(ya)

            # Normalize piercepoint locations to extent calculated above
            if not is_image_plane:
                x *= extent[0]
                y *= extent[1]

            if min_val is None:
                vmin = np.min([np.amin(screen), np.amin(fitted_phase1)])
            else:
                vmin = min_val
            if max_val is None:
                vmax = np.max([np.amax(screen), np.amax(fitted_phase1)])
            else:
                vmax = max_val

            logging.info('Plotting screens...')
            mpm = multiprocManager(ncpu, _plot_frame)
            for k in range(N_times):
                mpm.put([screen[:, :, k], fitted_phase1[:, k], residuals[:, k],
                weights[:, k, sindx], x[k, :], y[k, :], k, lower, upper, vmin, vmax,
                source_names, show_source_names, station_names, sindx, root_dir,
                prestr, is_image_plane, midRA, midDec, station_order[0, k, sindx], is_phase])
            mpm.wait()
    else:
        logging.info('Calculating screen images...')
        residuals = inresiduals.transpose([0, 2, 1]).reshape(N_piercepoints, N_times)
        weights = weights.transpose([0, 2, 1]).reshape(N_piercepoints, N_times)
        mpm = multiprocManager(ncpu, _calculate_screen)
        for k in range(N_times):
            mpm.put([inscreen[:, k, :], residuals[:, k], pp[k, :, :],
                N_piercepoints, k, east, north, up, T, Nx, Ny, -1, height,
                beta_val, r_0, is_phase])
        mpm.wait()
        for (k, ft, scr, xa, ya) in mpm.get():
            screen[:, :, k] = scr
            fitted_phase1[:, k] = ft
            if is_image_plane:
                x[k, :] = xa
                y[k, :] = ya
            else:
                x[k, :] = xa - np.amin(xa) # remove offsets for each time slot
                y[k, :] = ya - np.amin(ya)

        # Normalize piercepoint locations to extent calculated above
        if not is_image_plane:
            x *= extent[0]
            y *= extent[1]

        if min_val is None:
            vmin = np.min([np.amin(screen), np.amin(fitted_phase1)])
        else:
            vmin = min_val
        if max_val is None:
            vmax = np.max([np.amax(screen), np.amax(fitted_phase1)])
        else:
            vmax = max_val

        logging.info('Plotting screens...')
        if show_source_names:
            pp_names = []
            for stat in station_names:
                for src in source_names:
                    pp_names.append('{0}_{1}'.format(src, stat))
        else:
            pp_names = source_names
        mpm = multiprocManager(ncpu, _plot_frame)
        for k in range(N_times):
            order = station_order[0]
            mpm.put([screen[:, :, k], fitted_phase1[:, k], residuals[:, k],
            weights[:, k], x[k, :], y[k, :], k, lower, upper, vmin, vmax,
            pp_names, show_source_names, station_names, -1, root_dir,
            prestr, is_image_plane, midRA, midDec, order, is_phase])
        mpm.wait()


def _fitPLaneLTSQ(XYZ):
    """
    Fits a plane to an XYZ point cloud

    Returns (a, b, c), where Z = aX + bY + c

    Parameters
    ----------
    XYZ : array
        point cloud

    Returns
    -------
    (a, b, c) : floats
        plane parameters, where Z = aX + bY + c

    """
    import numpy as np
    [rows, cols] = XYZ.shape
    G = np.ones((rows, 3))
    G[:, 0] = XYZ[:, 0]  #X
    G[:, 1] = XYZ[:, 1]  #Y
    Z = XYZ[:, 2]
    (a, b, c), resid, rank, s = np.linalg.lstsq(G, Z)
    return (a, b, c)


def run(soltab, resSoltab='', minZ=-3.2, maxZ=3.2, prefix='', remove_gradient=False,
    show_source_names=False, ncpu=0):
    """
    Plot screens (one plot is made per time and per station)

    Parameters
    ----------
    soltab : solution table
        Soltab containing screen.
    resSoltab : solution table, optional
        Soltab containing the screen residuals.
    minZ : float, optional
        Minimum value of colorbar scale.
    maxZ : float, optional
        Max value of colorbar scale.
    prefix : str, optional
        String to prepend to output plots.
    remove_gradient : bool, optional
        If True, remove gradient before plotting.
    show_source_names : bool, optional
        If True, source names are overplotted.
    ncpu : int, optional
        Number of CPUs to use. If 0, all are used.

    """
    import os
    import numpy as np

    screen_type = soltab.getType()
    if screen_type == 'phasescreen':
        is_phase = True
    else:
        is_phase = False
    logging.info('Using input {0} soltab {1}'.format(screen_type, soltab.name))

    # Get values from soltabs
    solset = soltab.getSolset()
    if resSoltab is '':
        try:
            # Look for residual soltab assuming standard naming conventions
            ressoltab = solset.getSoltab(soltab.name+'resid')
        except:
            logging.error('Could not find the soltab with associated screen residuals. '
                'Please specify it with the "resSoltab" argument.')
            return 1
    else:
        ressoltab = solset.getSoltab(resSoltab)
    logging.info('Using input screen residual soltab: {}'.format(ressoltab.name))
    screen = np.array(soltab.val)
    weights = np.array(soltab.weight)
    residuals = np.array(ressoltab.val)
    times = np.array(soltab.time)
    orders = np.array(ressoltab.weight)
    axis_names = soltab.getAxesNames()
    if len(screen.shape) > 3:
        # remove degenerate freq and pol axes
        if 'freq' in axis_names:
            freq_ind = axis_names.index('freq')
            screen = np.squeeze(screen, axis=freq_ind)
            weights = np.squeeze(weights, axis=freq_ind)
            residuals = np.squeeze(residuals, axis=freq_ind)
            orders = np.squeeze(orders, axis=freq_ind)
            axis_names.remove('freq')
        if 'pol' in axis_names:
            pol_ind = axis_names.index('pol')
            screen = np.squeeze(screen, axis=pol_ind)
            weights = np.squeeze(weights, axis=pol_ind)
            residuals = np.squeeze(residuals, axis=pol_ind)
            orders = np.squeeze(orders, axis=pol_ind)
            axis_names.remove('pol')

    # Rearrange to get order [dir, time, ant]
    dir_ind = axis_names.index('dir')
    time_ind = axis_names.index('time')
    ant_ind = axis_names.index('ant')
    screen = screen.transpose([dir_ind, time_ind, ant_ind])
    weights = weights.transpose([dir_ind, time_ind, ant_ind])
    residuals = residuals.transpose([dir_ind, time_ind, ant_ind])
    orders = orders.transpose([dir_ind, time_ind, ant_ind])

    # Collect station and source names and positions and times, making sure
    # that they are ordered correctly.
    source_names = soltab.dir[:]
    source_dict = solset.getSou()
    source_positions = []
    for source in source_names:
        source_positions.append(source_dict[source])
    station_names = soltab.ant
    station_dict = solset.getAnt()
    station_positions = []
    for station in station_names:
        station_positions.append(station_dict[station])
    height = soltab.obj._v_attrs['height']
    beta_val = soltab.obj._v_attrs['beta']
    r_0 = soltab.obj._v_attrs['r_0']
    pp = soltab.obj.piercepoint[:]
    if height == 0.0:
        midRA = soltab.obj._v_attrs['midra']
        midDec = soltab.obj._v_attrs['middec']
    else:
        midRA = 0.0
        midDec = 0.0

    if (minZ == 0 and maxZ == 0):
        min_val = None
        max_val = None
    else:
        min_val = minZ
        max_val = maxZ

    _make_screen_plots(pp, screen, residuals, weights, np.array(station_names),
        np.array(station_positions), np.array(source_names), times,
        height, orders, beta_val, r_0, prefix=prefix,
        remove_gradient=remove_gradient, show_source_names=show_source_names, min_val=min_val,
        max_val=max_val, is_phase=is_phase, midRA=midRA, midDec=midDec, ncpu=ncpu)

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Some utilities for operations

import os, sys, ast, re
import functools, warnings # for deprecated_alias decorator
from configparser import ConfigParser
if (sys.version_info > (3, 0)):
    #from configparser import ConfigParser
    from io import StringIO
else:
    #from ConfigParser import ConfigParser
    from StringIO import StringIO

import numpy as np
from losoto._logging import logger as logging

cacheSteps = ['plot','clip','flag','norm','smooth'] # steps to use chaced data

class LosotoParser(ConfigParser):
    """
    A parser for losoto parset files.

    Parameters
    ----------
    parsetFile : str
        Name of the parset file.
    """

    def __init__(self, parsetFile):
        ConfigParser.__init__(self, inline_comment_prefixes=('#',';'))

        config = StringIO()
        # add [_global] fake section at beginning
        config.write('[_global]\n'+open(parsetFile).read())
        config.seek(0, os.SEEK_SET)
        self.readfp(config)

    def checkSpelling(self, s, soltab, availValues=[]):
        """
        check if any value in the step is missing from a value list and return a warning
        """
        entries = [x.lower() for x in list(dict(self.items(s)).keys())]
        availValues = ['soltab','operation'] + availValues + \
                    soltab.getAxesNames() + [a+'.minmaxstep' for a in soltab.getAxesNames()] + [a+'.regexp' for a in soltab.getAxesNames()]
        availValues = [x.lower() for x in availValues]
        for e in entries:
            if e not in availValues:
                logging.warning('Mispelled option: %s - Ignoring!' % e)

    def getstr(self, s, v, default=None):
        if self.has_option(s, v):
            return str(self.get(s, v).replace('\'','').replace('"','')) # remove apex
        elif default is None:
            logging.error('Section: %s - Values: %s: required (expected string).' % (s, v))
        else:
            return default

    def getbool(self, s, v, default=None):
        if self.has_option(s, v):
            return ConfigParser.getboolean(self, s, v)
        elif default is None:
            logging.error('Section: %s - Values: %s: required (expected bool).' % (s, v))
        else:
            return default

    def getfloat(self, s, v, default=None):
        if self.has_option(s, v):
            return ConfigParser.getfloat(self, s, v)
        elif default is None:
            logging.error('Section: %s - Values: %s: required (expected float).' % (s, v))
        else:
            return default

    def getint(self, s, v, default=None):
        if self.has_option(s, v):
            return ConfigParser.getint(self, s, v)
        elif default is None:
            logging.error('Section: %s - Values: %s: required (expected int).' % (s, v))
        else:
            return default

    def getarray(self, s, v, default=None):
        if self.has_option(s, v):
            try:
                # TODO: why are square brackets being replaced here? What if my selection is a string containing square brackets?
                # return self.getstr(s, v).replace(' ','').replace('[','').replace(']','').split(',') # split also turns str into 1-element lists
                parm = self.getstr(s, v) # split also turns str into 1-element lists
                if parm[0] == '[': # hardcoded for square brackets in parameter set...
                    parm = parm[1:]
                if parm[-1] == ']':
                    parm = parm[:-1]
                return parm.replace(' ','').split(',')
            except:
                logging.error('Error interpreting section: %s - values: %s (should be a list as [xxx,yyy,zzz...])' % (s, v))
        elif default is None:
            logging.error('Section: %s - Values: %s: required.' % (s, v))
        else:
            return default

    def getarraystr(self, s, v, default=None):
        try:
            return [str(x) for x in self.getarray(s, v, default)]
        except:
            logging.error('Error interpreting section: %s - values: %s (expected array of str.)' % (s, v))

    def getarraybool(self, s, v, default=None):
        try:
            return [bool(x) for x in self.getarray(s, v, default)]
        except:
            logging.error('Error interpreting section: %s - values: %s (expected array of bool.)' % (s, v))

    def getarrayfloat(self, s, v, default=None):
        try:
            return [float(x) for x in self.getarray(s, v, default)]
        except:
            logging.error('Error interpreting section: %s - values: %s (expected array of float.)' % (s, v))

    def getarrayint(self, s, v, default=None):
        try:
            return [int(x) for x in self.getarray(s, v, default)]
        except:
            logging.error('Error interpreting section: %s - values: %s (expected array of int.)' % (s, v))

    def getarrayfloat2d(self, s, v, default=None):
        """Alternative to parse 1,2 or ndim array input.
           'getarrayfloat() does not support more than 1 dim.
           TODO: it might be cleaner to unify these functions..."""
        try:
            # Remove space after [
            x = self.getstr(s, v, default)
            x = re.sub('\[ +', '[', x.strip())
            # Replace commas and spaces
            x = re.sub('[,\s]+', ', ', x)
            return np.array(ast.literal_eval(x))
        except:
            logging.error('Error interpreting section: %s - values: %s (expected array of float.)' % (s, v))


def getParAxis( parser, step, axisName ):
    """
    Axes values can be selected with:
    - an array of values
    - a string (reg exp): passed as axisName.regexp = ...
    - a min,max,step format: passed as axisName.minmaxstep = ...

    Parameters
    ----------
    parser : parser obj
        configuration file

    step : str
        this step

    axisName : str
        an axis name

    Returns
    -------
    str, dict or list
        a selection criteria
    """
    axisOpt = None

    if parser.has_option(step, axisName):
        axisOpt = parser.getarray(step, axisName)
    elif parser.has_option(step, axisName+'.regexp'):
        axisOpt = parser.getstr(step, axisName+'.regexp')
    elif parser.has_option(step, axisName+'.minmaxstep'):
        axisOpt = parser.getarray(step, axisName+'.minmaxstep')
        if len(axisOpt) == 3:
            axisOpt = {'min':float(axisOpt[0]), 'max':float(axisOpt[1]), 'step':int(axisOpt[2])}
        else:
            axisOpt = {'min':float(axisOpt[0]), 'max':float(axisOpt[1])}

    # global options
    elif parser.has_option('_global', axisName):
        axisOpt = parser.getarray('_global', axisName)
    elif parser.has_option('_global', axisName+'.regexp'):
        axisOpt = parser.getstr('_global', axisName+'.regexp')
    elif parser.has_option('_global', axisName+'.minmaxstep'):
        axisOpt = parser.getarray('_global', axisName+'.minmaxstep')
        if len(axisOpt) == 3:
            axisOpt = {'min':float(axisOpt[0]), 'max':float(axisOpt[1]), 'step':int(axisOpt[2])}
        else:
            axisOpt = {'min':float(axisOpt[0]), 'max':float(axisOpt[1])}

    if axisOpt == '' or axisOpt == []:
        axisOpt = None

    return axisOpt


def getStepSoltabs(parser, step, H):
    """
    Return a list of soltabs object for a step and apply selection creteria

    Parameters
    ----------
    parser : parser obj
        configuration file

    step : str
        current step

    H : h5parm obj
        the h5parm object

    Returns
    -------
    list
        list of soltab obj with applied selection
    """

    # selection on soltabs
    if parser.has_option(step, 'soltab'):
        stsel = parser.getarraystr(step, 'soltab')
    elif parser.has_option('_global', 'soltab'):
        stsel = parser.getarraystr('_global', 'soltab')
    else:
        stsel = ['.*/.*'] # select all
    #if not type(stsel) is list: stsel = [stsel]

    soltabs = []
    for solset in H.getSolsets():
        for soltabName in solset.getSoltabNames():
            if any(re.compile(this_stsel).match(solset.name+'/'+soltabName) for this_stsel in stsel):
                if parser.getstr(step, 'operation').lower() in cacheSteps:
                    soltabs.append( solset.getSoltab(soltabName, useCache=True) )
                else:
                    soltabs.append( solset.getSoltab(soltabName, useCache=False) )

    if soltabs == []:
        logging.warning('No soltabs selected for step %s.' % step)

    # axes selection
    for soltab in soltabs:
        userSel = {}
        for axisName in soltab.getAxesNames():
            userSel[axisName] = getParAxis( parser, step, axisName )
        soltab.setSelection(**userSel)

    return soltabs

# fancy backwards compatibility of keywords: allow aliases
# https://stackoverflow.com/questions/49802412/how-to-implement-deprecation-in-python-with-argument-alias#
def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)
        return wrapper
    return deco

def rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError('{} received both {} and {}'.format(
                    func_name, alias, new))
            warnings.warn('{} is deprecated; use {}'.format(alias, new),
                          DeprecationWarning)
            kwargs[new] = kwargs.pop(alias)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading FLAGEXTEND module.')

def _run_parser(soltab, parser, step):
    axesToExt = parser.getarraystr( step, 'axesToExt') # no default
    size = parser.getarrayint( step, 'size' ) # no default
    percent = parser.getfloat( step, 'percent', 50. )
    maxCycles = parser.getint( step, 'maxCycles', 3 )
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['axesToExt', 'size', 'percent', 'maxCycles'])
    return run(soltab, axesToExt, size, percent, maxCycles, ncpu)


def _flag(weights, coord, axesToExt, selection, percent=50, size=[0], maxCycles=3, outQueue=None):
        """
        Flag data if surreounded by other flagged data
        weights = the weights to convert into flags
        percent = percent of surrounding flagged point to extend the flag
        
        return: flags array and final rms
        """
        def extendFlag(flags, percent):
            #flags = flags.astype(np.int)
            if float(np.sum( flags ))/len(flags) > percent/100.:
                return 1
            else:
                return 0

        def percentFlagged(w):
            return 100.*(weights.size-np.count_nonzero(weights))/float(weights.size)


        import scipy.ndimage
        import numpy as np
        initPercent = percentFlagged(weights)

        # if size=0 then extend to all 2*axis, this otherwise create issues with mirroring
        for i, s in enumerate(size):
            if s == 0: size[i] = 2*weights.shape[i]

        for cycle in range(maxCycles):
            flag = scipy.ndimage.filters.generic_filter((weights==0), extendFlag, size=size, mode='mirror', cval=0.0, origin=0, extra_keywords={'percent':percent})
            weights[ ( flag == 1 ) ] = 0
            # no new flags
            if cycle != 0 and np.count_nonzero(flag) == oldFlagCount: break
            oldFlagCount = np.count_nonzero(flag)

        if percentFlagged(weights) == initPercent:
            logging.debug('Percentage of data flagged (%s): %.3f -> None' \
                    % (removeKeys(coord, axesToExt), initPercent))
        else:
            logging.debug('Percentage of data flagged (%s): %.3f -> %.3f %%' \
                    % (removeKeys(coord, axesToExt), initPercent, percentFlagged(weights)))

        outQueue.put([weights, selection])
        
            
def run( soltab, axesToExt, size, percent=50., maxCycles=3, ncpu=0 ):
    """
    This operation for LoSoTo implement a extend flag procedure
    It can work in multi dimensional space and for each datum check if the surrounding data are flagged to a certain %, then flag also that datum
    The size of the surrounding footprint can be tuned
    WEIGHT: compliant

    Parameters
    ----------
    axesToExt : list of str
        Axes used to find close flags.

    size : list of int
        Size of the window (diameter), per axis. If 0 is given then the entire length of the axis is assumed.
        Must be a vector of same length of Axes.

    percent : float, optional
        Percent of flagged data around the point to flag it, by default 50.

    maxCycles : int, optional
        Number of independent cycles of flag expansion, by default 3.

    ncpu : int, optional
        Number of CPU used, by default all available.
    """

    import numpy as np

    logging.info("Extending flag on soltab: "+soltab.name)

    # input check
    if axesToExt == []:
        logging.error("Please specify at least one axis to extend flag.")
        return 1

    # start processes for multi-thread
    mpm = multiprocManager(ncpu, _flag)

    for axisToExt in axesToExt:
        if axisToExt not in soltab.getAxesNames():
            logging.error('Axis \"'+axisToExt+'\" not found.')
            mpm.wait()
            return 1

    # fill the queue (note that sf and sw cannot be put into a queue since they have file references)
    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=axesToExt, weight=True):
        mpm.put([weights, coord, axesToExt, selection, percent, size, maxCycles])

    mpm.wait()

    logging.info('Writing solutions')
    for w, sel in mpm.get():
        soltab.setValues(w, sel, weight=True)

    soltab.addHistory('FLAG EXTENDED (over %s)' % (str(axesToExt)))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading ABS module.')

def _run_parser(soltab, parser, step):
    parser.checkSpelling( step, soltab )
    return run(soltab)

def run( soltab ):
    """
    Take absolute value. Needed before smooth if amplitudes are negative!
    WEIGHT: no need to be weight compliant

    Parameters
    ----------
    soltab : soltab obj
        Solution table.
    """

    import numpy as np

    logging.info("Taking ABSolute value of soltab: "+soltab.name)

    vals = soltab.getValues(retAxesVals = False)
    count = np.count_nonzero(vals<0)

    logging.info('Abs: %i points initially negative (%f %%)' % (count,100*float(count)/np.count_nonzero(vals)))

    # writing back the solutions
    soltab.setValues(np.abs(vals))

    soltab.addHistory('ABSolute value taken')
        
    return 0





import os, sys, logging, time

class _ColorStreamHandler(logging.StreamHandler):

    DEFAULT = '\x1b[0m'
    RED     = '\x1b[31m'
    GREEN   = '\x1b[32m'
    YELLOW  = '\x1b[33m'
    PURP    = '\x1b[35m'

    CRITICAL = RED
    ERROR    = RED
    WARNING  = YELLOW
    INFO     = GREEN
    DEBUG    = PURP

    @classmethod
    def _get_color(cls, level):
        if level >= logging.CRITICAL:  return cls.CRITICAL
        elif level >= logging.ERROR:   return cls.ERROR
        elif level >= logging.WARNING: return cls.WARNING
        elif level >= logging.INFO:    return cls.INFO
        elif level >= logging.DEBUG:   return cls.DEBUG
        else:                          return cls.DEFAULT

    def __init__(self, stream=None):
        logging.StreamHandler.__init__(self, stream)

    def format(self, record):
        color = self._get_color(record.levelno)
        record.msg = color + record.msg + self.DEFAULT
        return logging.StreamHandler.format(self, record)

class Logger():

    def __init__(self, level = 'info', logfile = None, log_dir = None):

        self.logfile = logfile
        self.log_dir = log_dir
        self.backup(logfile, log_dir)
        self.set_logger(logfile, log_dir)
        self.set_level(level)

    def backup(self, logfile, log_dir):

        if self.logfile is not None:
            # bkp old log dir
            if os.path.isdir(log_dir):
                current_time = time.localtime()
                log_dir_old = time.strftime(log_dir+'_bkp_%Y-%m-%d_%H:%M', current_time)
                os.system('mv %s %s' % ( log_dir, log_dir_old ))
            os.makedirs(log_dir)
    
            # bkp old log file
            if os.path.exists(logfile):
                current_time = time.localtime()
                logfile_old = time.strftime(logfile+'_bkp_%Y-%m-%d_%H:%M', current_time)
                os.system('mv %s %s' % ( logfile, logfile_old ))
            

    def set_logger(self, logfile, log_dir):
      
        self.logger = logging.getLogger("LoSoTo")
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S")

        # create file handler which logs even debug messages
        if self.logfile is not None:
            handlerFile = logging.FileHandler(logfile)
            #handlerFile.setLevel(logging.DEBUG)
            handlerFile.setFormatter(formatter)
            self.logger.addHandler(handlerFile)
        
        # create console handler with a higher log level
        handlerConsole = _ColorStreamHandler(stream=sys.stdout)
        #handlerConsole.setLevel(logging.INFO)
        handlerConsole.setFormatter(formatter)
        self.logger.addHandler(handlerConsole)

    def set_level(self, level):
        if level == 'warning':
            self.logger.setLevel(logging.WARNING)
        elif level == 'info':
            self.logger.setLevel(logging.INFO)
        elif level == 'debug':
            self.logger.setLevel(logging.DEBUG)
        else:
            print("Debug level %s doesn't exist." % level)

# this is used by all libraries for logging
logger = logging.getLogger("LoSoTo")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading CLIP module.')

def _run_parser(soltab, parser, step):
    axesToClip = parser.getarraystr( step, 'axesToClip', [] )
    clipLevel = parser.getfloat( step, 'clipLevel', 5. )
    log = parser.getbool( step, 'log', False )
    mode = parser.getstr( step, 'mode', 'median' )
    
    parser.checkSpelling( step, soltab, ['axesToClip', 'clipLevel', 'log', 'mode'] )
    return run(soltab, axesToClip, clipLevel, log, mode)

def run( soltab, axesToClip=None, clipLevel=5., log=False, mode='median' ):
    """
    Clip solutions around the median by a factor specified by the user.
    WEIGHT: flag compliant, putting weights into median is tricky

    Parameters
    ----------
    axesToClip : list of str
        axes along which to calculate the median (e.g. [time,freq]).

    clipLevel : float, optional
        factor above/below median at which to clip, by default 5.

    log : bool, optional
        clip is done in log10 space, by default False.

    mode : str, optional
        if "median" then flag at rms*clipLevel times away from the median, if it is "above", 
        then flag all values above clipLevel, if it is "below" then flag all values below clipLevel. 
        By default median.
    """

    import numpy as np

    def percentFlagged(w):
        return 100.*(weights.size-np.count_nonzero(weights))/float(weights.size)

    logging.info("Clipping soltab: "+soltab.name)

    # input check
    if len(axesToClip) < 1 and mode == 'median':
        logging.error("Please specify axes to clip.")
        return 1
    else:
        axesToClip = soltab.getAxesNames()

    if mode != 'median' and mode != 'above' and mode != 'below':
        logging.error("Mode can be only: median, above or below.")
        return 1

    if mode == 'median' and soltab.getType() == 'amplitude' and not log:
        logging.warning('Amplitude solution tab detected and log=False. Amplitude solution tables should be treated in log space.')

    # some checks
    for i, axis in enumerate(axesToClip[:]):
        if axis not in soltab.getAxesNames():
            del axesToClip[i]
            logging.warning('Axis \"'+axis+'\" not found. Ignoring.')

    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=axesToClip, weight = True):

        initPercent = percentFlagged(weights)

        # skip all flagged
        if (weights == 0).all():
            continue
        
        if log: vals = np.log10(vals)

        if mode == 'median':
            valmedian = np.nanmedian(vals[(weights != 0)])
            rms = np.nanstd(vals[(weights != 0)])
            np.putmask(weights, np.abs(vals-valmedian) > rms * clipLevel, 0)

        elif mode == 'above':
            np.putmask(weights, vals > clipLevel, 0)

        elif mode == 'below':
            np.putmask(weights, vals < clipLevel, 0)
    
        # writing back the solutions
        if log: vals = 10**vals
        soltab.setValues(weights, selection, weight=True)

        #print('max', np.max(vals[(weights != 0)]))
        #print('median', np.nanmedian(vals[(weights != 0)]))
        logging.debug('Percentage of data flagged (%s): %.3f%% -> %.3f%%' \
            % (removeKeys(coord, axesToClip), initPercent, percentFlagged(weights)))

    soltab.addHistory('CLIP (over %s with %s sigma cut)' % (axesToClip, clipLevel))

    soltab.flush()
        
    return 0





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This is the station-screen operation for LoSoTo

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading STATIONSCREEN module.')

def _run_parser(soltab, parser, step):
    outSoltab = parser.getstr( step, "outSoltab" )
    order = parser.getint( step, "Order", 5 )
    beta = parser.getfloat( step, "Beta", 5.0/3.0 )
    niter = parser.getint( step, "niter", 2 )
    nsigma = parser.getfloat( step, "nsigma", 5.0 )
    refAnt = parser.getint( step, "RefAnt", -1 )
    scale_order = parser.getbool( step, "ScaleOrder", True )
    scale_dist = parser.getfloat( step, "scaleDist", 25000.0 )
    min_order = parser.getint( step, "MinOrder", 5 )
    adjust_order = parser.getbool( step, "AdjustOrder", True )
    ncpu = parser.getint( step, "ncpu", 0 )

    parser.checkSpelling( step, soltab, ['outSoltab', 'order', 'beta', 'niter', 'nsigma',\
        'refAnt', 'scale_order', 'scale_dist', 'min_order', 'adjust_order'])
    return run(soltab, outSoltab, order, beta, niter, nsigma,
        refAnt, scale_order, scale_dist, min_order, adjust_order, ncpu)


def _calculate_piercepoints(station_positions, source_positions):
    """
    Returns array of piercepoint locations

    Parameters
    ----------
    station_positions : array
        Array of station positions
    source_positions : array
        Array of source positions

    Returns
    -------
    pp : array
        Array of pierce points
    midRA : float
        Reference RA for WCS system (deg)
    midDec : float
        Reference Dec for WCS system (deg)

    """
    import numpy as np

    logging.info('Calculating screen pierce-point locations...')
    N_sources = source_positions.shape[0]
    N_stations = station_positions.shape[0]
    N_piercepoints = N_stations * N_sources
    pp = np.zeros((N_piercepoints, 3))

    xyz = np.zeros((N_sources, 3))
    ra_deg = source_positions.T[0] * 180.0 / np.pi
    dec_deg = source_positions.T[1] * 180.0 / np.pi
    xy, midRA, midDec = _getxy(ra_deg, dec_deg)
    xyz[:, 0] = xy[0]
    xyz[:, 1] = xy[1]
    pp_idx = 0
    for i in range(N_sources):
        for station_position in station_positions:
            pp[pp_idx, :] = xyz[i]
            pp_idx += 1

    return pp, midRA, midDec


def _get_ant_dist(ant_xyz, ref_xyz):
    """
    Returns distance between ant and ref in m

     Parameters
    ----------
    ant_xyz : array
        Array of station position
    ref_xyz : array
        Array of reference position

    Returns
    -------
    dist : float
        Distance between station and reference positions

    """
    import numpy as np

    return np.sqrt((ref_xyz[0] - ant_xyz[0])**2 + (ref_xyz[1] - ant_xyz[1])**2 + (ref_xyz[2] - ant_xyz[2])**2)


def _getxy(RA, Dec, midRA=None, midDec=None):
    """
    Returns array of projected x and y values.

    Parameters
    ----------
    RA, Dec : list
        Lists of RA and Dec in degrees
    midRA : float
        RA for WCS reference in degrees
    midDec : float
        Dec for WCS reference in degrees

    Returns
    -------
    x, y : numpy array, numpy array, float, float
        arrays of x and y values

    """
    import numpy as np

    if midRA is None or midDec is None:
        x, y  = _radec2xy(RA, Dec)

        # Refine x and y using midpoint
        if len(x) > 1:
            xmid = min(x) + (max(x) - min(x)) / 2.0
            ymid = min(y) + (max(y) - min(y)) / 2.0
            xind = np.argsort(x)
            yind = np.argsort(y)
            try:
                midxind = np.where(np.array(x)[xind] > xmid)[0][0]
                midyind = np.where(np.array(y)[yind] > ymid)[0][0]
                midRA = RA[xind[midxind]]
                midDec = Dec[yind[midyind]]
                x, y  = _radec2xy(RA, Dec, midRA, midDec)
            except IndexError:
                midRA = RA[0]
                midDec = Dec[0]
        else:
            midRA = RA[0]
            midDec = Dec[0]

    x, y  = _radec2xy(RA, Dec, refRA=midRA, refDec=midDec)

    return np.array([x, y]), midRA, midDec


def _radec2xy(RA, Dec, refRA=None, refDec=None):
    """
    Returns x, y for input ra, dec.

    Note that the reference RA and Dec must be the same in calls to both
    _radec2xy() and _xy2radec() if matched pairs of (x, y) <=> (RA, Dec) are
    desired.

    Parameters
    ----------
    RA : list
        List of RA values in degrees
    Dec : list
        List of Dec values in degrees
    refRA : float, optional
        Reference RA in degrees.
    refDec : float, optional
        Reference Dec in degrees

    Returns
    -------
    x, y : list, list
        Lists of x and y pixel values corresponding to the input RA and Dec
        values

    """
    import numpy as np

    x = []
    y = []
    if refRA is None:
        refRA = RA[0]
    if refDec is None:
        refDec = Dec[0]

    # Make wcs object to handle transformation from ra and dec to pixel coords.
    w = _makeWCS(refRA, refDec)

    for ra_deg, dec_deg in zip(RA, Dec):
        ra_dec = np.array([[ra_deg, dec_deg]])
        x.append(w.wcs_world2pix(ra_dec, 0)[0][0])
        y.append(w.wcs_world2pix(ra_dec, 0)[0][1])

    return x, y


def _xy2radec(x, y, refRA=0.0, refDec=0.0):
    """
    Returns x, y for input ra, dec.

    Note that the reference RA and Dec must be the same in calls to both
    _radec2xy() and _xy2radec() if matched pairs of (x, y) <=> (RA, Dec) are
    desired.

    Parameters
    ----------
    x : list
        List of x values in pixels
    y : list
        List of y values in pixels
    refRA : float, optional
        Reference RA in degrees
    refDec : float, optional
        Reference Dec in degrees

    Returns
    -------
    RA, Dec : list, list
        Lists of RA and Dec values corresponding to the input x and y pixel
        values

    """
    import numpy as np

    RA = []
    Dec = []

    # Make wcs object to handle transformation from ra and dec to pixel coords.
    w = _makeWCS(refRA, refDec)

    for xp, yp in zip(x, y):
        x_y = np.array([[xp, yp]])
        RA.append(w.wcs_pix2world(x_y, 0)[0][0])
        Dec.append(w.wcs_pix2world(x_y, 0)[0][1])

    return RA, Dec


def _makeWCS(refRA, refDec):
    """
    Makes simple WCS object.

    Parameters
    ----------
    refRA : float
        Reference RA in degrees
    refDec : float
        Reference Dec in degrees

    Returns
    -------
    w : astropy.wcs.WCS object
        A simple TAN-projection WCS object for specified reference position

    """
    from astropy.wcs import WCS
    import numpy as np

    w = WCS(naxis=2)
    w.wcs.crpix = [1000, 1000]
    w.wcs.cdelt = np.array([-0.0005, 0.0005])
    w.wcs.crval = [refRA, refDec]
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    w.wcs.set_pv([(2, 1, 45.0)])

    return w


def _flag_outliers(weights, residual, nsigma, screen_type):
    """
    Flags outliers

    Parameters
    ----------
    weights : array
        Array of weights
    phase_residual : array
        Array of residual values from phase screen fitting (rad)
    nsigma : float
        Number of sigma above with outliers are clipped (= weight set to zero)
    screen_type : str
        Type of screen: 'phase', 'tec', or 'amplitude'

    Returns
    -------
    weights : array
        array of weights, with flagged times set to 0


    """
    import numpy as np
    from losoto.operations.reweight import _nancircstd

    # Find stddev of the screen
    flagged = np.where(weights == 0.0)
    nonflagged = np.where(weights > 0.0)
    if nonflagged[0].size == 0:
        return weights

    if screen_type == 'phase':
        # Use circular stddev
        residual = normalize_phase(residual)
        residual_nan = residual.copy()
        residual_nan[flagged] = np.nan
        screen_stddev = _nancircstd(residual_nan, axis=0)
    elif screen_type == 'tec' or screen_type == 'amplitude':
        # Use normal stddev
        screen_stddev = np.sqrt(np.average(residual[nonflagged]**2,
            weights=weights[nonflagged], axis=0))

    # Compare residuals to stddev of the screen
    outlier_ind = np.where(np.abs(residual) > nsigma*screen_stddev)
    weights[outlier_ind] = 0.0

    return weights


def _circ_chi2(samples, weights):
    """
    Compute the circular chi^2

    Based on scipy.stats.circstd

    Parameters
    ----------
    samples : array_like
        Input array.
    weights : array_like
        Input array.

    Returns
    -------
    chi2 : float
        Circular chi^2.
    """
    import numpy as np

    unflagged = np.where(weights > 0.0)
    if unflagged[0].size == 0:
        return 0.0

    x1 = np.sin(samples[unflagged])
    x2 = np.cos(samples[unflagged])
    meanx1, sumw = np.average(x1**2, weights=weights[unflagged], returned=True)
    meanx2, sumw = np.average(x2**2, weights=weights[unflagged], returned=True)
    R = np.hypot(meanx1, meanx2)
    var = (1.0 - R)

    return var * sumw


def _calculate_svd(pp, r_0, beta, N_piercepoints):
    """
    Returns result (U) of svd for K-L vectors

    Parameters
    ----------
    pp : array
        Array of piercepoint locations
    r_0: float
        Scale size of amp fluctuations (m)
    beta: float
        Power-law index for amp structure function (5/3 => pure Kolmogorov
        turbulence)
    N_piercepoints : int
        Number of piercepoints

    Returns
    -------
    C : array
        C matrix
    pinvC : array
        Inv(C) matrix
    U : array
        Unitary matrix

    """
    import numpy as np
    from scipy.linalg import pinv, svd

    D = np.resize(pp, (N_piercepoints, N_piercepoints, 3))
    D = np.transpose(D, (1, 0, 2)) - D
    D2 = np.sum(D**2, axis=2)
    C = -(D2 / r_0**2)**(beta / 2.0) / 2.0
    pinvC = pinv(C, rcond=1e-3)
    U, S, V = svd(C)

    return C, pinvC, U


def _fit_screen(source_names, full_matrices, pp, rr, weights, order, r_0, beta,
    screen_type):
    """
    Fits a screen to amplitudes or phases using Karhunen-Lo`eve base vectors

    Parameters
    ----------
    source_names: array
        Array of source names
    full_matrices : list of arrays
        List of [C, pivC, U] matrices for all piercepoints
    pp: array
        Array of piercepoint locations
    airmass: array
        Array of airmass values (note: not currently used)
    rr: array
        Array of amp values to fit screen to
    weights: array
        Array of weights
    order: int
        Order of screen (i.e., number of KL base vectors to keep)
    r_0: float
        Scale size of amp fluctuations (m)
    beta: float
        Power-law index for amp structure function (5/3 => pure Kolmogorov
        turbulence)
    screen_type : str
        Type of screen: 'phase' or 'amplitude'

    Returns
    -------
    screen_fit_white_all, screen_residual_all : array, array
        Arrays of screen and residual (actual - screen) values

    """
    import numpy as np
    from scipy.linalg import pinv

    # Identify flagged directions
    N_sources_all = len(source_names)
    unflagged = np.where(weights > 0.0)
    N_sources = len(source_names[unflagged])

    # Initialize arrays
    N_piercepoints = N_sources
    N_piercepoints_all = N_sources_all
    screen_fit_all = np.zeros((N_sources_all, 1))
    pp_all = pp.copy()
    rr_all = rr.copy()
    pp = pp_all[unflagged[0], :]
    w = np.diag(weights[unflagged])

    # Calculate matrices
    if N_sources == N_sources_all:
        C, pinvC, U = full_matrices
    else:
        # Recalculate for unflagged directions
        C, pinvC, U = _calculate_svd(pp, r_0, beta, N_piercepoints)
    invU = pinv(np.dot(np.transpose(U[:, :order]), np.dot(w, U)[:, :order]), rcond=1e-3)

    # Fit screen to unflagged directions
    if screen_type == 'phase':
        # Change phase to real/imag
        rr_real = np.cos(rr[unflagged])
        rr_imag = np.sin(rr[unflagged])

        # Calculate real screen
        rr1 = np.dot(np.transpose(U[:, :order]), np.dot(w, rr_real))
        real_fit = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))

        # Calculate imag screen
        rr1 = np.dot(np.transpose(U[:, :order]), np.dot(w, rr_imag))
        imag_fit = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))

        # Calculate phase screen
        screen_fit = np.arctan2(np.dot(C, imag_fit), np.dot(C, real_fit))
        screen_fit_white = np.dot(pinvC, screen_fit)
    elif screen_type == 'amplitude':
        # Calculate log(amp) screen
        rr = rr[unflagged]
        rr1 = np.dot(np.transpose(U[:, :order]), np.dot(w, np.log10(rr)))
        amp_fit_log = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))

        # Calculate amp screen
        screen_fit = np.dot(C, amp_fit_log)
        screen_fit_white = np.dot(pinvC, screen_fit)
    elif screen_type == 'tec':
        # Calculate tec screen
        rr = rr[unflagged]
        rr1 = np.dot(np.transpose(U[:, :order]), np.dot(w, rr))
        tec_fit = np.dot(pinvC, np.dot(U[:, :order], np.dot(invU, rr1)))

        # Calculate amp screen
        screen_fit = np.dot(C, tec_fit)
        screen_fit_white = np.dot(pinvC, screen_fit)

    # Calculate screen in all directions
    if N_sources != N_sources_all:
        screen_fit_all[unflagged[0], :] = screen_fit[:, np.newaxis]
        flagged = np.where(weights <= 0.0)
        for findx in flagged[0]:
            p = pp_all[findx, :]
            d2 = np.sum(np.square(pp - p), axis=1)
            c = -(d2 / ( r_0**2 ))**(beta / 2.0) / 2.0
            screen_fit_all[findx, :] = np.dot(c, screen_fit_white)
        C, pinvC, U = full_matrices
        screen_fit_white_all = np.dot(pinvC, screen_fit_all)
        if screen_type == 'amplitude':
            screen_residual_all = rr_all - 10**(screen_fit_all.reshape(N_piercepoints_all))
        else:
            screen_residual_all = rr_all - screen_fit_all.reshape(N_piercepoints_all)
    else:
        screen_fit_white_all = screen_fit_white
        if screen_type == 'amplitude':
            screen_residual_all = rr_all - 10**(np.dot(C, screen_fit_white))
        else:
            screen_residual_all = rr_all - np.dot(C, screen_fit_white)
    screen_fit_white_all = screen_fit_white_all.reshape((N_sources_all, 1))
    screen_residual_all = screen_residual_all.reshape((N_sources_all, 1))

    return (screen_fit_white_all, screen_residual_all)

def _process_station(rr, pp, screen_order, station_weights, screen_type, niter, nsigma,
                     adjust_order, source_names, full_matrix, beta, r_0):
    """
    Processes a station

    Parameters
    ----------
    soltab: solution table
        Soltab containing amplitude solutions
    outsoltab: str
        Name of output soltab
    order : int, optional
        Order of screen (i.e., number of KL base vectors to keep). If the order
        is scaled by dist (scale_order = True), the order is calculated as
        order * sqrt(dist/scale_dist)
    beta: float, optional
        Power-law index for amp structure function (5/3 => pure Kolmogorov
        turbulence)
    niter: int, optional
        Number of iterations to do when determining weights
    nsigma: float, optional
        Number of sigma above which directions are flagged
    refAnt: str or int, optional
        Index (if int) or name (if str) of reference station (-1 => no ref)
    scale_order : bool, optional
        If True, scale the screen order with sqrt of distance/scale_dist to the
        reference station
    scale_dist : float, optional
        Distance used to normalize the distances used to scale the screen order.
        If None, the max distance is used
    adjust_order : bool, optional
        If True, adjust the screen order to obtain a reduced chi^2 of approx.
        unity
    min_order : int, optional
        The minimum allowed order if adjust_order = True.

    """
    # Iterate:
    # 1. fit screens
    # 2. flag nsigma outliers
    # 3. refit with new weights
    # 4. repeat for niter
    target_redchi2 = 1.0
    N_sources = rr.shape[0]
    N_times = rr.shape[1]
    screen = np.zeros((N_sources, N_times))
    residual = np.zeros((N_sources, N_times))
    station_order = screen_order[0]
    init_station_weights = station_weights.copy() # preserve initial weights
    for iterindx in range(niter):
        if iterindx > 0:
            # Flag outliers
            if screen_type == 'phase' or screen_type == 'tec':
                # Use residuals
                screen_diff = residual.copy()
            elif screen_type == 'amplitude':
                # Use log residuals
                screen_diff = np.log10(rr) - np.log10(np.abs(rr - residual))
            station_weights = _flag_outliers(init_station_weights,
                    screen_diff, nsigma, screen_type)

        # Fit the screens
        norderiter = 1
        if adjust_order:
            if iterindx > 0:
                norderiter = 4
        for tindx in range(N_times):
            N_unflagged = np.where(station_weights[:, tindx] > 0.0)[0].size
            if N_unflagged == 0:
                continue
            if screen_order[tindx] > N_unflagged-1:
                screen_order[tindx] = N_unflagged-1
            hit_upper = False
            hit_lower = False
            hit_upper2 = False
            hit_lower2 = False
            sign = 1.0
            for oindx in range(norderiter):
                skip_fit = False
                if iterindx > 0:
                    if np.all(station_weights[:, tindx] == prev_station_weights[:, tindx]):
                        if not adjust_order:
                            # stop fitting if weights did not change
                            break
                        elif oindx == 0:
                            # Skip the fit for first iteration, as it is the same as the prev one
                            skip_fit = True
                if not np.all(station_weights[:, tindx] == 0.0) and not skip_fit:
                    scr, res = _fit_screen(source_names, full_matrix, pp[:, :], rr[:, tindx],
                                           station_weights[:, tindx],
                                           int(screen_order[tindx]), r_0,
                                           beta, screen_type)
                    screen[:, tindx] = scr[:, 0]
                    residual[:, tindx] = res[:, 0]

                if hit_lower2 or hit_upper2:
                    break

                if adjust_order and iterindx > 0:
                    if screen_type == 'phase':
                        redchi2 =  _circ_chi2(residual[:, tindx],
                            station_weights[:, tindx]) / (N_unflagged - screen_order[tindx])
                    elif screen_type == 'amplitude':
                        # Use log residuals
                        screen_diff = np.log10(rr[:, tindx]) - np.log10(np.abs(rr[:, tindx] - residual[:, tindx]))
                        redchi2 = np.sum(np.square(screen_diff) *
                                         station_weights[:, tindx]) / (N_unflagged - screen_order[tindx])
                    else:
                        redchi2 = np.sum(np.square(residual[:, tindx]) *
                            station_weights[:, tindx]) / (N_unflagged - screen_order[tindx])
                    if oindx > 0:
                        if redchi2 > 1.0 and prev_redchi2 < redchi2:
                            sign *= -1
                        if redchi2 < 1.0 and prev_redchi2 > redchi2:
                            sign *= -1
                    prev_redchi2 = redchi2
                    order_factor = (N_unflagged - screen_order[tindx])**0.2
                    target_order = float(screen_order[tindx]) - sign * order_factor * (target_redchi2 - redchi2)
                    target_order = max(station_order, target_order)
                    target_order = min(int(round(target_order)), N_unflagged-1)
                    if target_order <= 0:
                        target_order = min(station_order, N_unflagged-1)
                    if target_order == screen_order[tindx]:# don't fit again if order is the same as last one
                        break
                    if target_order == N_unflagged-1:# check whether we've been here before. If so, break
                        if hit_upper:
                            hit_upper2 = True
                        hit_upper = True
                    if target_order == station_order:# check whether we've been here before. If so, break
                        if hit_lower:
                            hit_lower2 = True
                        hit_lower = True
                    screen_order[tindx] = target_order
        prev_station_weights = station_weights.copy()

    return screen, station_weights, residual, screen_order


def _process_single_freq(freq_ind, screen_type, niter, nsigma, refAnt,
                         adjust_order, source_names, station_names, beta, r_0, outQueue):
    global r_full, weights_full, screen, residual, full_matrices, screen_order, pp

    N_sources, N_stations, N_times, N_freqs, N_pols = screen.shape
    weights_out = np.zeros(weights_full[:, :, 0, :, :].shape)
    screen_out = np.zeros(screen[:, :, :, 0, :].shape)
    residual_out = np.zeros(residual[:, :, :, 0, :].shape)
    screen_order_out = np.zeros(screen_order[:, :, 0, :].shape)
    for pol_ind in range(N_pols):
        r = r_full[:, :, freq_ind, :, pol_ind] # order is now [dir, time, ant]
        r = r.transpose([0, 2, 1]) # order is now [dir, ant, time]
        weights = weights_full[:, :, freq_ind, :, pol_ind]
        weights = weights.transpose([0, 2, 1])

        # Fit screens
        for s, stat in enumerate(station_names):
            if s == refAnt and (screen_type == 'phase' or screen_type == 'tec'):
                # skip reference station (phase- or tec-type only)
                continue
            if np.all(np.isnan(r[:, s, :])) or np.all(weights[:, s, :] == 0):
                # skip fully flagged stations
                continue
            rr = np.reshape(r[:, s, :], [N_sources, N_times])
            station_weights = weights[:, s, :]
            station_screen_order = screen_order[s, :, freq_ind, pol_ind]
            scr, wgt, res, sord = _process_station(rr, pp, station_screen_order, station_weights,
                                             screen_type, niter, nsigma, adjust_order,
                                             source_names, full_matrices, beta, r_0)
            screen_out[:, s, :, pol_ind] = scr
            weights[:, s, :] = wgt
            residual_out[:, s, :, pol_ind] = res
            screen_order_out[s, :, pol_ind] = sord
        weights_out[:, :, :, pol_ind] = weights.transpose([0, 2, 1]) # order is now [dir, time, ant]

    outQueue.put([freq_ind, screen_out, weights_out, residual_out, screen_order_out])


def run(soltab, outsoltab, order=12, beta=5.0/3.0, niter=2, nsigma=5.0,
    refAnt=-1, scale_order=True, scale_dist=None, min_order=5, adjust_order=True, ncpu=0):
    """
    Fits station screens to input soltab (type 'phase' or 'amplitude' only).

    The results of the fit are stored in the soltab parent solset in "outsoltab"
    and the residual values (actual - screen) are stored in "outsoltabresid".
    These values are the screen amplitude values per station per pierce point
    per solution interval. The pierce point locations are stored in an auxiliary
    array in the output soltabs.

    Screens can be plotted with the PLOTSCREEN operation.

    Parameters
    ----------
    soltab: solution table
        Soltab containing amplitude solutions
    outsoltab: str
        Name of output soltab
    order : int, optional
        Order of screen (i.e., number of KL base vectors to keep). If the order
        is scaled by dist (scale_order = True), the order is calculated as
        order * sqrt(dist/scale_dist)
    beta: float, optional
        Power-law index for amp structure function (5/3 => pure Kolmogorov
        turbulence)
    niter: int, optional
        Number of iterations to do when determining weights
    nsigma: float, optional
        Number of sigma above which directions are flagged
    refAnt: str or int, optional
        Index (if int) or name (if str) of reference station (-1 => no ref)
    scale_order : bool, optional
        If True, scale the screen order with sqrt of distance/scale_dist to the
        reference station
    scale_dist : float, optional
        Distance used to normalize the distances used to scale the screen order.
        If None, the max distance is used
    adjust_order : bool, optional
        If True, adjust the screen order to obtain a reduced chi^2 of approx.
        unity
    min_order : int, optional
        The minimum allowed order if adjust_order = True.
    ncpu : int, optional
        Number of CPUs to use. If 0, all are used

    """
    import numpy as np
    global r_full, weights_full, screen, residual, full_matrices, screen_order, pp

    # Get screen type
    screen_type = soltab.getType()
    if screen_type not in ['phase', 'amplitude', 'tec']:
        logging.error('Screens can only be fit to soltabs of type "phase", "tec", or "amplitude".')
        return 1
    logging.info('Using solution table {0} to calculate {1} screens'.format(soltab.name, screen_type))

    # Load values, etc.
    r_full = np.array(soltab.val)
    weights_full = soltab.weight[:]
    times = np.array(soltab.time)
    freqs = soltab.freq[:]
    axis_names = soltab.getAxesNames()
    freq_ind = axis_names.index('freq')
    dir_ind = axis_names.index('dir')
    time_ind = axis_names.index('time')
    ant_ind = axis_names.index('ant')
    if 'pol' in axis_names:
        is_scalar = False
        pol_ind = axis_names.index('pol')
        N_pols = len(soltab.pol[:])
        r_full = r_full.transpose([dir_ind, time_ind, freq_ind, ant_ind, pol_ind])
        weights_full = weights_full.transpose([dir_ind, time_ind, freq_ind, ant_ind, pol_ind])
    else:
        is_scalar = True
        N_pols = 1
        r_full = r_full.transpose([dir_ind, time_ind, freq_ind, ant_ind])
        r_full = r_full[:, :, :, :, np.newaxis]
        weights_full = weights_full.transpose([dir_ind, time_ind, freq_ind, ant_ind])
        weights_full = weights_full[:, :, :, :, np.newaxis]

    # Collect station and source names and positions and times, making sure
    # that they are ordered correctly.
    solset = soltab.getSolset()
    source_names = soltab.dir[:]
    source_dict = solset.getSou()
    source_positions = []
    for source in source_names:
        source_positions.append(source_dict[source])
    station_names = soltab.ant[:]
    if type(station_names) is not list:
        station_names = station_names.tolist()
    station_dict = solset.getAnt()
    station_positions = []
    for station in station_names:
        station_positions.append(station_dict[station])
    N_sources = len(source_names)
    N_times = len(times)
    N_stations = len(station_names)
    N_freqs = len(freqs)
    N_piercepoints = N_sources

    # Set ref station and reference phases if needed
    if type(refAnt) is str:
        if N_stations == 1:
            refAnt = -1
        elif refAnt in station_names:
            refAnt = station_names.index(refAnt)
        else:
            refAnt = -1
    if refAnt != -1 and screen_type == 'phase' or screen_type == 'tec':
        r_ref = r_full[:, :, :, refAnt, :].copy()
        for i in range(len(station_names)):
            r_full[:, :, :, i, :] -= r_ref

    if scale_order:
        dist = []
        if refAnt == -1:
            station_order = [order] * N_stations
        else:
            for s in range(len(station_names)):
                dist.append(_get_ant_dist(station_positions[s], station_positions[refAnt]))
            if scale_dist is None:
                scale_dist = max(dist)
            logging.info('Using variable order (with max order = {0} '
                'and scaling dist = {1} m)'.format(order, scale_dist))
            station_order = []
            for s in range(len(station_names)):
                station_order.append(max(min_order, min(order, int(order * np.sqrt(dist[s] / scale_dist)))))
    else:
        station_order = [order] * len(station_names)
        logging.info('Using order = {0}'.format(order))

    # Initialize various arrays and parameters
    screen = np.zeros((N_sources, N_stations, N_times, N_freqs, N_pols))
    residual = np.zeros((N_sources, N_stations, N_times, N_freqs, N_pols))
    screen_order = np.zeros((N_stations, N_times, N_freqs, N_pols))
    for s in range(len(station_names)):
        for freq_ind in range(N_freqs):
            for pol_ind in range(N_pols):
                screen_order[s, :, freq_ind, pol_ind] = station_order[s]
    r_0 = 100

    # Calculate full piercepoint arrays
    pp, midRA, midDec = _calculate_piercepoints(np.array([station_positions[0]]),
                                                np.array(source_positions))
    full_matrices = _calculate_svd(pp, r_0, beta, N_piercepoints)

    # Fit station screens
    mpm = multiprocManager(ncpu, _process_single_freq)
    for freq_ind in range(N_freqs):
         mpm.put([freq_ind, screen_type, niter, nsigma, refAnt,
                  adjust_order, source_names, station_names, beta, r_0])
    mpm.wait()
    for (freq_ind, scr, wgt, res, sord) in mpm.get():
        screen[:, :, :, freq_ind, :] = scr
        residual[:, :, :, freq_ind, :] = res
        weights_full[:, :, freq_ind, :, :] = wgt
        screen_order[:, :, freq_ind, :] = sord

    # Write the results to the output solset
    dirs_out = source_names
    times_out = times
    ants_out = station_names
    freqs_out = freqs

    # Store screen values
    # Note: amplitude screens are log10() values with non-log residuals!
    vals = screen.transpose([2, 3, 1, 0, 4]) # order is now ['time', 'freq', 'ant', 'dir', 'pol']
    weights = weights_full.transpose([1, 2, 3, 0, 4]) # order is now ['time', 'freq', 'ant', 'dir', 'pol']
    if is_scalar:
        screen_st = solset.makeSoltab('{}screen'.format(screen_type), outsoltab,
            axesNames=['time', 'freq', 'ant', 'dir'], axesVals=[times_out, freqs_out,
            ants_out, dirs_out], vals=vals[:, :, :, :, 0], weights=weights[:, :, :, :, 0])
        vals = residual.transpose([2, 3, 1, 0, 4])
        weights = np.zeros(vals.shape)
        for d in range(N_sources):
            # Store the screen order as the weights of the residual soltab
            weights[:, :, :, d, :] = screen_order.transpose([1, 2, 0, 3]) # order is now [time, ant, freq, pol]
        resscreen_st = solset.makeSoltab('{}screenresid'.format(screen_type), outsoltab+'resid',
            axesNames=['time', 'freq', 'ant', 'dir'], axesVals=[times_out, freqs_out,
            ants_out, dirs_out], vals=vals[:, :, :, :, 0], weights=weights[:, :, :, :, 0])
    else:
        pols_out = soltab.pol[:]
        screen_st = solset.makeSoltab('{}screen'.format(screen_type), outsoltab,
            axesNames=['time', 'freq', 'ant', 'dir', 'pol'], axesVals=[times_out, freqs_out,
            ants_out, dirs_out, pols_out], vals=vals, weights=weights)
        vals = residual.transpose([2, 3, 1, 0, 4])
        weights = np.zeros(vals.shape)
        for d in range(N_sources):
            # Store the screen order as the weights of the residual soltab
            weights[:, :, :, d, :] = screen_order.transpose([1, 2, 0, 3]) # order is now [time, ant, freq, pol]
        resscreen_st = solset.makeSoltab('{}screenresid'.format(screen_type), outsoltab+'resid',
            axesNames=['time', 'freq', 'ant', 'dir', 'pol'], axesVals=[times_out, freqs_out,
            ants_out, dirs_out, pols_out], vals=vals, weights=weights)

    # Store beta, r_0, height, and order as attributes of the screen soltabs
    screen_st.obj._v_attrs['beta'] = beta
    screen_st.obj._v_attrs['r_0'] = r_0
    screen_st.obj._v_attrs['height'] = 0.0
    screen_st.obj._v_attrs['midra'] = midRA
    screen_st.obj._v_attrs['middec'] = midDec

    # Store piercepoint table. Note that it does not conform to the axis
    # shapes, so we cannot use makeSoltab()
    solset.obj._v_file.create_array('/'+solset.name+'/'+screen_st.obj._v_name,
        'piercepoint', obj=pp)

    screen_st.addHistory('CREATE (by STATIONSCREEN operation)')
    resscreen_st.addHistory('CREATE (by STATIONSCREEN operation)')

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading NORM module.')

def _run_parser(soltab, parser, step):
    axesToNorm = parser.getarraystr( step, 'axesToNorm' ) # no default
    normVal = parser.getfloat( step, 'normVal', 1.)
    log = parser.getbool( step, 'log', False )

    parser.checkSpelling( step, soltab, ['axesToNorm','normVal','log'])
    return run(soltab, axesToNorm, normVal, log)

def run( soltab, axesToNorm, normVal = 1., log = False ):
    """
    Normalize the solutions to a given value
    WEIGHT: Weights compliant

    Parameters
    ----------
    axesToNorm : array of str
        Axes along which compute the normalization.

    normVal : float, optional
        Number to normalize to vals = vals * (normVal/valsMean), by default 1.

    log : bool, optional
        clip is done in log10 space, by default False.
    """
    import numpy as np
    
    logging.info("Normalizing soltab: "+soltab.name)

    # input check
    axesNames = soltab.getAxesNames()
    for normAxis in axesToNorm:
        if normAxis not in axesNames:
            logging.error('Normalization axis '+normAxis+' not found.')
            return 1

    if soltab.getType() == 'amplitude' and not log:
        logging.warning('Amplitude solution tab detected and log=False. Amplitude solution tables should be treated in log space.')

    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=axesToNorm, weight = True):

        if log: vals = np.log10(vals)

        # rescale solutions
        if np.all(weights == 0): continue # skip flagged selections
        valsMean = np.nanmean(vals[weights != 0])

        if log: vals[weights != 0] += np.log10(normVal)-valsMean
        else: vals[weights != 0] *= normVal/valsMean

        logging.debug("Rescaling by: "+str(normVal/valsMean))

        # writing back the solutions
        if log: vals = 10**vals
        soltab.setValues(vals, selection)

    soltab.flush()
    soltab.addHistory('NORM (on axis %s)' % (axesToNorm))

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading PREFACTOR_XYOFFSET module.')

def _run_parser(soltab, parser, step):
    chanWidth = parser.getstr( step, 'chanWidth')

    parser.checkSpelling( step, soltab, ['chanWidth'])
    return run(soltab, chanWidth)


def normalize(phase):
    """
    Normalize phase to the range [-pi, pi].
    """

    # Convert to range [-2*pi, 2*pi].
    out = np.fmod(phase, 2.0 * np.pi)

    # Convert to range [-pi, pi]
    out[out < -np.pi] += 2.0 * np.pi
    out[out > np.pi] -= 2.0 * np.pi

    return out


def run( soltab, chanWidth ):

    import numpy as np
    import scipy.signal as sg

    logging.info("Running XYoffset on: "+soltab.name)
    solset = soltab.getSolset()

    solType = soltab.getType()
    if solType != 'phase':
       logging.error("Soltab type of "+soltab.name+" is: "+solType+" but should be phase.")
       return 1

    for vals, coord, selection in soltab.getValuesIter(returnAxes=['pol','ant','freq', 'time'], weight=False):
        vals = reorderAxes( vals, soltab.getAxesNames(), ['time', 'ant', 'freq', 'pol'] )
        pass

    phases_tmp = np.copy(vals) # axes are [time, ant, freq, pol]
    freqs = np.copy(soltab.freq)
    npol = len(soltab.pol)
    refstationID=2

    stationsnames = [ stat for stat in soltab.ant]

    subbandHz = 195.3125e3
    if type(chanWidth) is str:
        letters = [1 for s in chanWidth[::-1] if s.isalpha()]
        indx = len(chanWidth) - sum(letters)
        unit = chanWidth[indx:]
        if unit.strip().lower() == 'hz':
            conversion = 1.0
        elif unit.strip().lower() == 'khz':
            conversion = 1e3
        elif unit.strip().lower() == 'mhz':
            conversion = 1e6
        else:
            logging.error("The unit on chanWidth was not understood.")
            raise ValueError("The unit on chanWidth was not understood.")
        chanWidthHz = float(chanWidth[:indx]) * conversion
    else:
        chanWidthHz = chanWidth
    offsetHz = subbandHz / 2.0 - 0.5 * chanWidthHz
    freqmin = np.min(soltab.freq[:]) + offsetHz # central frequency of first subband
    freqmax = np.max(soltab.freq[:]) + offsetHz # central frequency of last subband
    freqs_new  = np.arange(freqmin, freqmax+100e3, subbandHz)

    # this gets the subband number to any given frequency in HBA-low
    subbands = np.unique(np.round(freqs/195.3125e3-512.))
    nsubbands = len(subbands)
    nchan = len(freqs)/nsubbands
    if nsubbands*nchan != len(freqs):
        print("find_cal_global_phaseoffset_losoto.py: irregular number of ch/SB detected! Bailing out!")
        print("  nchan %d, nSB: %d, nfreq: %d" % (nchan, nsubbands, len(freqs)))
        sys.exit(1)
    tmpfreqs = freqs.reshape([nsubbands,nchan])
    freq_per_sb = np.mean(tmpfreqs,axis=1)
    nstations = len(stationsnames)
    refphases = phases_tmp[:, refstationID, :, :]

    for istat in range(nstations):
        phases_00 = phases_tmp[:, istat, :, 0] - refphases[:, :, 0]
        phases_11 = phases_tmp[:, istat, :, 1] - refphases[:, :, 1]
        phases_diff = normalize(phases_00 - phases_11)
        tmp_phases_diff = np.median(phases_diff, axis=0)  # take median over time axis
        med_phases_diff = np.median(tmp_phases_diff.reshape([nsubbands, nchan]), axis=1)  # take median over each subband
        if istat == 0:
            global_stat_offsets = med_phases_diff
        else:
            global_stat_offsets = np.vstack( (global_stat_offsets, med_phases_diff) )
    global_stat_offsets_smoothed = np.zeros([nsubbands, nstations, npol])
    global_stat_offsets_smoothed_interp = np.zeros([len(freqs_new), nstations, npol])
    for istat in range(nstations):
        global_stat_offsets_smoothed[:, istat, -1] = sg.medfilt(global_stat_offsets[istat, :], kernel_size=15) # smooth over frequency

        # Convert to real/imag, invert correction (so that offsets are removed when applied),
        # and interpolate to the output frequency grid
        real = np.interp(freqs_new, freq_per_sb, np.cos(-1. * global_stat_offsets_smoothed[:, istat, -1]))
        imag = np.interp(freqs_new, freq_per_sb, np.sin(-1. * global_stat_offsets_smoothed[:, istat, -1]))
        global_stat_offsets_smoothed_interp[:, istat, -1] = np.arctan2(imag, real)

    try:
        new_soltab = solset.getSoltab('XYoffset')
        new_soltab.delete()
    except:
        pass
    new_soltab = solset.makeSoltab(soltype='phase', soltabName='XYoffset',
                             axesNames=['freq', 'ant', 'pol'], axesVals=[freqs_new, soltab.ant, ['XX','YY']],
                             vals=global_stat_offsets_smoothed_interp,
                             weights=np.ones_like(global_stat_offsets_smoothed_interp, dtype=np.float16))
    new_soltab.addHistory('CREATE (by PREFACTOR_XYOFFSET operation)')

    return 0





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This tool is used to convert an H5parm file to parmdb format by writing to
# existing parmdb instrument table(s).
#
# It handles Gain/DirectionalGain/RotationAngle/CommonRotationAngle/CommonScalarPhase solution types.
_author = "Francesco de Gasperin (astro@voo.it), David Rafferty (drafferty@hs.uni-hamburg.de)"

import sys, os, glob, re, time
import numpy as np
import shutil
import pyrap.tables as pt
import lofar.parmdb
from losoto import _version
from losoto import _logging
from losoto.h5parm import h5parm
try:
    import progressbar
except ImportError:
    import losoto.progressbar as progressbar


def parmdbToAxes(solEntry):
    """
    Extract the information written as a string in the parmdb format
    """
    pol = None
    pol1 = None
    pol2 = None
    dir = None
    ant = None
    parm = None

    thisSolType = solEntry.split(':')[0]

    # For CommonRotationAngle assuming [CommonRotationAngle:ant]
    if thisSolType == 'CommonRotationAngle':
        thisSolType, ant = solEntry.split(':')
        dir = 'pointing'

    # For RotationAngle assuming [RotationAngle:ant:sou]
    elif thisSolType == 'RotationAngle':
        thisSolType, ant, dir = solEntry.split(':')

    # For CommonScalarPhase assuming [CommonScalarPhase:ant]
    elif thisSolType == 'CommonScalarPhase':
        thisSolType, ant = solEntry.split(':')
        dir = 'pointing'

    # For CommonScalarPhase assuming [CommonScalarPhase:ant]
    elif thisSolType == 'CommonScalarAmplitude':
        thisSolType, ant = solEntry.split(':')
        dir = 'pointing'

    # For ScalarPhase assuming [ScalarPhase:ant:sou]
    elif thisSolType == 'ScalarPhase':
        thisSolType, ant, dir = solEntry.split(':')

    # For ScalarPhase assuming [ScalarPhase:ant:sou]
    elif thisSolType == 'ScalarAmplitude':
        thisSolType, ant, dir = solEntry.split(':')

    # For TEC assuming [TEC:ant or TEC:pol:ant]
    elif thisSolType == 'TEC':
        try:
            thisSolType, ant = solEntry.split(':')
        except:
            thisSolType, pol, ant = solEntry.split(':')
            pol1 = pol
            pol2 = pol
        dir = 'pointing'

    # For Clock assuming [Clock:ant or Clock:pol:ant]
    elif thisSolType == 'Clock':
        try:
            thisSolType, ant = solEntry.split(':')
        except:
            thisSolType, pol, ant = solEntry.split(':')
            pol1 = pol
            pol2 = pol
        dir = 'pointing'

    # For RotationMeasure assuming [RotationMeasure:ant]
    elif thisSolType == 'RotationMeasure':
        dir = 'pointing'
        try:
            thisSolType, ant = solEntry.split(':')
        except:
            thisSolType, ant, dir = solEntry.split(':')

    # For Gain assuming [Gain:pol1:pol2:parm:ant]
    elif thisSolType == 'Gain':
        thisSolType, pol1, pol2, parm, ant = solEntry.split(':')
        dir = 'pointing'

    # For DirectionalGain assuming [DirecitonalGain:pol1:pol2:parm:ant:sou]
    elif thisSolType == 'DirectionalGain':
        thisSolType, pol1, pol2, parm, ant, dir = solEntry.split(':')

    else:
        logging.error('Unknown solution type "'+thisSolType+'". Ignored.')

    if pol1 is not None and pol2 is not None:
        if pol1 == '0' and pol2 == '0': pol = 'XX'
        if pol1 == '1' and pol2 == '0': pol = 'YX'
        if pol1 == '0' and pol2 == '1': pol = 'XY'
        if pol1 == '1' and pol2 == '1': pol = 'YY'

    if pol is not None:
        pol = re.escape(pol)
    if dir is not None:
        dir = re.escape(dir)
    if ant is not None:
        ant = re.escape(ant)
    return pol, dir, ant, parm


def getSoltabFromSolType(solType, solTabs, parm='ampl'):
    """Return a list of solution tables that corresponds to the type given

    solType - string defining solution type. E.g., "DirectionalGain"
    solTabs - solution tables returned by h5parm.getSoltabs()
    parm - parm to return if solType is "Gain": 'ampl'/'real' or
           'phase'/'imag'

    The soltab parmdb_type attribute is used as an additional filter to
    to distinguish multiple possible matches. If it is not available, all
    matches are returned.
    """
    solTabList = []
    if parm is not None:
        parm = parm.lower()

    for st in solTabs:
        # Handle gain table separately, as we need to distinguish ampl and phase
        if solType == 'DirectionalGain' or solType == 'Gain':
            if (parm == 'ampl' or parm == 'real') and st.getType() == 'amplitude':
                if hasattr(st.obj._v_attrs, 'parmdb_type'):
                    if st.obj._v_attrs['parmdb_type'] is not None:
                        if solType in st.obj._v_attrs['parmdb_type'].split(', '):
                            solTabList.append(st)
                    else:
                        solTabList.append(st)
                else:
                    solTabList.append(st)
            elif (parm == 'phase' or parm == 'imag') and st.getType() == 'phase':
                if hasattr(st.obj._v_attrs, 'parmdb_type'):
                    if st.obj._v_attrs['parmdb_type'] is not None:
                        if solType in st.obj._v_attrs['parmdb_type'].split(', '):
                            solTabList.append(st)
                    else:
                        solTabList.append(st)
                else:
                    solTabList.append(st)
        else:
            if hasattr(st.obj._v_attrs, 'parmdb_type'):
                if st.obj._v_attrs['parmdb_type'] is not None:
                    if solType in st.obj._v_attrs['parmdb_type'].split(', '):
                        solTabList.append(st)
                else:
                    if (solType == 'RotationAngle' or solType == 'CommonRotationAngle') and st.getType() == 'rotation':
                        solTabList.append(st)
                    elif (solType == 'CommonScalarPhase' or solType == 'ScalarPhase') and st.getType() == 'scalarphase':
                        solTabList.append(st)
                    elif (solType == 'CommonScalarAmplitude' or solType == 'ScalarAmplitude') and st.getType() == 'scalaramplitude':
                        solTabList.append(st)
                    elif solType == 'Clock' and st.getType() == 'clock':
                        solTabList.append(st)
                    elif solType == 'TEC' and st.getType() == 'tec':
                        solTabList.append(st)
                    elif solType == 'RotationMeasure' and st.getType() == 'rotationmeasure':
                        solTabList.append(st)
            else:
                if (solType == 'RotationAngle' or solType == 'CommonRotationAngle') and st.getType() == 'rotation':
                    solTabList.append(st)
                elif (solType == 'CommonScalarPhase' or solType == 'ScalarPhase') and st.getType() == 'scalarphase':
                    solTabList.append(st)
                elif (solType == 'CommonScalarAmplitude' or solType == 'ScalarAmplitude') and st.getType() == 'scalaramplitude':
                    solTabList.append(st)
                elif solType == 'Clock' and st.getType() == 'clock':
                    solTabList.append(st)
                elif solType == 'TEC' and st.getType() == 'tec':
                    solTabList.append(st)
                elif solType == 'RotationMeasure' and st.getType() == 'rotationmeasure':
                    solTabList.append(st)

    if len(solTabList) == 0:
        return None
    else:
        return solTabList


def makeTECparmdb(H, solset, TECsolTab, timewidths, freq, freqwidth):
    """Returns TEC screen parmdb parameters

    H - H5parm object
    solset - solution set with TEC screen parameters
    TECsolTab = solution table with tecscreen values
    timewidths - time widths of output parmdb
    freq - frequency of output parmdb
    freqwidth - frequency width of output parmdb
    """
    global ipbar, pbar

    solset = H.getSolset(solset)

    station_dict = solset.getAnt()
    station_names = list(station_dict.keys())
    station_positions = list(station_dict.values())

    source_dict = solset.getSou()
    source_names = list(source_dict.keys())
    source_positions = list(source_dict.values())

    tec_sf = solset.getSoltab(TECsolTab)
    tec_screen, axis_vals = tec_sf.getValues()
    times = axis_vals['time']
    beta = TECsolTab._v_attrs['beta']
    r_0 = TECsolTab._v_attrs['r_0']
    height = TECsolTab._v_attrs['height']
    order = TECsolTab._v_attrs['order']
    pp = tec_sf.t.piercepoint

    N_sources = len(source_names)
    N_times = len(times)
    N_freqs = 1
    N_stations = len(station_names)
    N_piercepoints = N_sources * N_stations

    freqs = freq
    freqwidths = freqwidth
    parms = {}
    v = {}
    v['times'] = times
    v['timewidths'] = timewidths
    v['freqs'] = freqs
    v['freqwidths'] = freqwidths

    for station_name in station_names:
        for source_name in source_names:

            v['values'] = np.zeros((N_times, N_freqs), dtype=np.double)
            parmname = 'Piercepoint:X:%s:%s' % (station_name, source_name)
            parms[parmname] = v.copy()

            v['values'] = np.zeros((N_times, N_freqs), dtype=np.double)
            parmname = 'Piercepoint:Y:%s:%s' % (station_name, source_name)
            parms[parmname] = v.copy()

            v['values'] = np.zeros((N_times, N_freqs), dtype=np.double)
            parmname = 'Piercepoint:Z:%s:%s' % (station_name, source_name)
            parms[parmname] = v.copy()

            v['values'] = np.zeros((N_times, N_freqs), dtype=np.double)
            parmname = 'TECfit_white:%s:%s' % (station_name, source_name)
            parms[parmname] = v.copy()

            v['values'] = np.zeros((N_times, N_freqs), dtype=np.double)
            parmname = 'TECfit_white:0:%s:%s' % (station_name, source_name)
            parms[parmname] = v.copy()

            v['values'] = np.zeros((N_times, N_freqs), dtype=np.double)
            parmname = 'TECfit_white:1:%s:%s' % (station_name, source_name)
            parms[parmname] = v.copy()

    for k in range(N_times):
        D = np.resize(pp[k, :, :], (N_piercepoints, N_piercepoints, 3))
        D = np.transpose(D, ( 1, 0, 2 )) - D
        D2 = np.sum(D**2, axis=2)
        C = -(D2 / (r_0**2))**(beta / 2.0) / 2.0
        tec_fit_white = np.dot(np.linalg.inv(C),
            tec_screen[:, k, :].reshape(N_piercepoints))
        pp_idx = 0
        for src, source_name in enumerate(source_names):
            for sta, station_name in enumerate(station_names):

                parmname = 'Piercepoint:X:%s:%s' % (station_name, source_name)
                parms[parmname]['values'][k, 0] = pp[k, pp_idx, 0]

                parmname = 'Piercepoint:Y:%s:%s' % (station_name, source_name)
                parms[parmname]['values'][k, 0] = pp[k, pp_idx, 1]

                parmname = 'Piercepoint:Z:%s:%s' % (station_name, source_name)
                parms[parmname]['values'][k, 0] = pp[k, pp_idx, 2]

                parmname = 'TECfit_white:%s:%s' % (station_name, source_name)
                parms[parmname]['values'][k, 0] = tec_fit_white[pp_idx]

                parmname = 'TECfit_white:0:%s:%s' % (station_name, source_name)
                parms[parmname]['values'][k, 0] = tec_fit_white[pp_idx]

                parmname = 'TECfit_white:1:%s:%s' % (station_name, source_name)
                parms[parmname]['values'][k, 0] = tec_fit_white[pp_idx]

                pp_idx += 1
        pbar.update(ipbar)
        ipbar += 1

    time_start = times[0] - timewidths[0]/2
    time_end = times[-1] + timewidths[-1]/2

    v['times'] = np.array([(time_start + time_end) / 2])
    v['timewidths'] = np.array([time_end - time_start])

    v_r0 = v.copy()
    v_r0['values'] = np.array(r_0, dtype=np.double, ndmin=2)
    parms['r_0'] = v_r0

    v_beta = v.copy()
    v_beta['values'] = np.array(beta, dtype=np.double, ndmin=2)
    parms['beta'] = v_beta

    v_height = v.copy()
    v_height['values'] = np.array(height, dtype=np.double, ndmin=2)
    parms['height'] = v_height

    return parms


if __name__=='__main__':
    # Options
    import optparse
    opt = optparse.OptionParser(usage='%prog <H5parm filename> <output globaldb/SB filename>\n'+
        _author, version='%prog '+_version.__version__)
    opt.add_option('-V', '--verbose', help='Go VeRbOsE!',
        action='store_true', default=False)
    opt.add_option('-s', '--solset', help='Name of solution set to export '
        '(default=sol000)', type='string', default='sol000')
    opt.add_option('-o', '--outfile', help='Filename of globaldb/SB to export parmdb to '
        '(default=input globaldb/SB filename)', type='string', default=None)
    opt.add_option('-r', '--root', help='Root string to prepend to input parmdb '
        'instrument directories to make the output parmdb directories '
        '(default=solution-set name)', type='string', default=None)
    opt.add_option('-t', '--soltab', help='Solution tables to export; e.g., '
        '"amplitude000, phase000" (default=all)', type='string', default='all')
    opt.add_option('-i', '--instrument', help='Name of the instrument table '
        '(default=instrument*)', type='string', default='instrument*')
    opt.add_option('-c', '--clobber', help='Clobber exising files '
        '(default=False)', action='store_true', default=False)
    (options, args) = opt.parse_args()
    global ipbar, pbar

    logger = _logging.Logger('info')
    logging = _logging.logger

    # Check options
    if len(args) != 2:
        opt.print_help()
        sys.exit()
    if options.verbose: logger.set_level("debug")

    # Check input H5parm file
    h5parmFile = args[0]
    if not os.path.exists(h5parmFile):
        logging.critical('Input H5parm file not found.')
        sys.exit(1)
    logging.info("Input H5parm filename = "+h5parmFile)

    # Open the h5parm file and get solution set names
    h5parm_in = h5parm(h5parmFile, readonly = True)
    solsetNames = h5parm_in.getSolsetNames()

    # Check input parmdb file
    globaldbFile = args[1]
    if not os.path.exists(globaldbFile):
        logging.critical('Input globaldb/SB file not found.')
        sys.exit(1)
    logging.info("Input globaldb/SB filename = "+globaldbFile)

    # Check input solution set name
    solsetName = options.solset
    if solsetName not in solsetNames:
        logging.critical('The solution set "'+solsetName+'" was not found in input H5parm file.')
        sys.exit(1)
    logging.info("Solution set name = "+solsetName)
    solset = h5parm_in.getSolset(solsetName)

    # Make output parmdb directory if needed
    out_globaldbFile = options.outfile
    if out_globaldbFile is None:
        out_globaldbFile = globaldbFile
    if not os.path.exists(out_globaldbFile):
        os.mkdir(out_globaldbFile)
    logging.info("Output globaldb/SB filename = "+out_globaldbFile)

    # Check output root
    outroot = options.root
    if outroot is None:
        outroot = solsetName

    # Make a list of all available instrument tables (only 1 for a standard MS)
    instrumentdbFiles = [ instrumentdbFile for instrumentdbFile in \
        glob.glob(os.path.join(globaldbFile, options.instrument)) \
        if os.path.isdir(instrumentdbFile) ]
    if len(instrumentdbFiles) == 0:
        logging.critical('No parmdb table(s) found in input globaldb/SB file.')
        sys.exit(1)
    instrumentdbFiles.sort()

    # Find solution table types using the first instrumentdb
    # TODO: is there a better solution which check all the instrumentdbs?
    pdb = lofar.parmdb.parmdb(instrumentdbFiles[0])
    solTypes = list(set(x[0] for x in  (x.split(":") for x in pdb.getNames())))
    solTabs = solset.getSoltabs()
    if options.soltab != 'all':
        soltabs_to_use = [s.strip() for s in options.soltab.split(',')]
        logging.info('Using solution tables: {0}'.format(soltabs_to_use))
        solTabs_filt = []
        for s in solTabs:
            if s.name in soltabs_to_use:
                solTabs_filt.append(s)
        for s in soltabs_to_use:
            if s.name not in solset.getSoltabNames():
                logging.warning('Solution table {0} not found in input H5parm file.'.format(s.name))
        solTabs = solTabs_filt
    if len(solTabs) == 0:
        logging.critical('No solution tables found in input H5parm file')
        sys.exit(1)
    pdbSolTypes = solTypes[:]
    for solType in pdbSolTypes:
        solTabList = getSoltabFromSolType(solType, solTabs, 'ampl')
        if solTabList is None:
            # Search for type phase solutions if no ampl ones where found
            solTabList = getSoltabFromSolType(solType, solTabs, 'phase')
        if solTabList is None:
            logging.warning("Solution type {0} not found in solution set {1}. Skipping.".format(solType, solsetName))
            solTypes.remove(solType)

    # Look for tecscreen solution table in the solset. If
    # found, add to solTypes
    st_tec = None
    for st in solTabs:
        if st.getType() == 'tecscreen':
            st_tec = st
    if st_tec is not None:
        solTypes.append('TECScreen')
    solTypes = list(set(solTypes))
    logging.info('Found solution types in input parmdb and H5parm: '+', '.join(solTypes))

    # For each solType, select appropriate solutions and construct
    # the dictionary to pass to pdb.addValues()
    len_sol = {}
    for solType in solTypes:
        if solType != 'TECScreen':
            len_sol[solType] = len(pdb.getNames(solType+':*'))
        else:
            N_times = st_tec.getAxisLen(axis='time')
            len_sol[solType] = N_times

    cachedSolTabs = {}
    for instrumentdbFile in instrumentdbFiles:
        out_instrumentdbFile = out_globaldbFile + '/' + outroot + '_' + instrumentdbFile.split('/')[-1]
        logging.info('Filling '+out_instrumentdbFile+':')

        # Remove existing instrumentdb (if clobber) and create new one
        if os.path.exists(out_instrumentdbFile):
            if options.clobber:
                shutil.rmtree(out_instrumentdbFile)
            else:
                logging.critical('Output instrumentdb file exists and '
                    'clobber = False.')
                sys.exit(1)
        pdb_out = lofar.parmdb.parmdb(out_instrumentdbFile+'/', create=True)

        pbar = progressbar.ProgressBar(maxval=sum(len_sol.values())).start()
        ipbar = 0

        pdb_in = lofar.parmdb.parmdb(instrumentdbFile)

        # Add default values and steps
        DefValues = pdb_in.getDefValues()
        for k, v in DefValues.items():
            pdb_out.addDefValues({k: pdb.makeDefValue(v.item(0))})
        pdb_out.setDefaultSteps(pdb_in.getDefaultSteps())

        for solType in solTypes:
            if len_sol[solType] == 0: continue

            if solType != 'TECScreen':
                solEntries = pdb_in.getNames(solType+':*')
                data = pdb_in.getValuesGrid(solType+':*')
                data_out = data.copy()
                for solEntry in solEntries:

                    pol, dir, ant, parm = parmdbToAxes(solEntry)
                    solTabList = getSoltabFromSolType(solType, solTabs, parm=parm)
                    if solTabList is None:
                        continue
                    if len(solTabList) > 1:
                        logging.warning('More than one solution table found in H5parm '
                                'matching parmdb entry "'+solType+'". Taking the first match: '+str(solTabList[0])+'.')
                    solTab = solTabList[0]

                    # search in the cache for open soltab
                    if not solTab.getType() in cachedSolTabs:
                        cachedSolTabs[solTab.getType()] = solTab
                    else:
                        solTab = cachedSolTabs[solTab.getType()]
                        solTab.setSelection()

                    freqs = data[solEntry]['freqs']
                    times = data[solEntry]['times']
                    parms = {}

                    if 'ant' in solTab.getAxesNames():
                        parms['ant'] = [ant]
                        # skip missing antennas (e.g. internationals sometimes are retained in the parmdb)
                        if not ant in solTab.getAxisValues('ant'): continue
                    if 'pol' in solTab.getAxesNames(): parms['pol'] = [pol]
                    if 'dir' in solTab.getAxesNames(): parms['dir'] = [dir]
                    if 'freq' in solTab.getAxesNames(): parms['freq'] = freqs.tolist()
                    # workaround for bbs and ndppp dealing differently with the last time slot when #timeslots%ntime != 0
                    # NDPPP has all intervals the same
                    # BBS has a maller interval in the last timeslot which is compensated here
                    if times[-1] - times[-2] < times[-2] - times[-3]: times[-1] = times[-2] + (times[-2] - times[-3])
                    if 'time' in solTab.getAxesNames(): parms['time'] = {'min':np.min(times-0.1), 'max':np.max(times+0.1)}
                    solTab.setSelection(**parms)

                    # If needed, convert Amp and Phase to Real and Imag
                    if parm == 'Real':
                        solTabList = getSoltabFromSolType(solType, solTabs, parm='phase')
                        if not solTabList[0].getType() in cachedSolTabs:
                            solTab_ph = solTabList[0]
                            cachedSolTabs[solTabList[0].getType()] = solTab_ph
                        else:
                            solTab_ph = cachedSolTabs[solTabList[0].getType()]
                        solTab_ph.setSelection(ant=[ant], pol=[pol], dir=[dir], freq=freqs.tolist(),
                            time={'min':np.min(times-0.1), 'max':np.max(times+0.1)})
                        val_phase = solTab_ph.getValues()[0]
                        val_amp = solTab.getValues()[0]
                        val = val_amp * np.cos(val_phase)
                    elif parm == 'Imag':
                        solTabList = getSoltabFromSolType(solType, solTabs, parm='ampl')
                        if not solTabList[0].getType() in cachedSolTabs:
                            solTab_amp = solTabList[0]
                            cachedSolTabs[solTabList[0].getType()] = solTab_amp
                        else:
                            solTab_amp = cachedSolTabs[solTabList[0].getType()]
                        solTab_amp.setSelection(ant=[ant], pol=[pol], dir=[dir], freq=freqs.tolist(),
                            time={'min':np.min(times-0.1), 'max':np.max(times+0.1)})
                        val_phase = solTab.getValues()[0]
                        val_amp = solTab_amp.getValues()[0]
                        val = val_amp * np.sin(val_phase)
                    else:
                        val = solTab.getValues()[0]

                    # Apply flags
                    weights = solTab.getValues(weight=True)[0]

                    # etienne part; if it is borken, curse his name
                    # check whether this is clock or tec; if so, reshape properly to account for all freqs in the parmdb
                    # anyway these tables are freq-indep
                    if solType == "Clock":# or solType == "TEC" or solType == "RotationMeasure":
                        # find freq-dimensionality 
                        nfreq = freqs.shape[0]
                        #print val.shape
                        # reshape such that all freq arrays are filled properly
                        val = np.tile( val, np.append([nfreq], np.ones(len(val.shape),dtype=np.int) ) )
                        #print val.shape
                        weights = np.tile( weights, np.append([nfreq], np.ones(len(weights.shape),dtype=np.int) ) )

                    flags = np.zeros(shape=weights.shape, dtype=bool)
                    flags[np.where(weights == 0)] = True
                    if parm == 'Real':
                        weights2 = solTab_ph.getValues(weight=True)[0]
                        flags[np.where(weights2 == 0)] = True
                    if parm == 'Imag':
                        weights2 = solTab_amp.getValues(weight=True)[0]
                        flags[np.where(weights2 == 0)] = True
                    np.putmask(val, flags, np.nan)

                    shape = data_out[solEntry]['values'].shape
                    #print "shape"
                    #print 'parmdb', shape
                    #print 'h5parm', val.shape
                    #print sf.getAxesNames()
                    #print sf.getType()
                    #print "parmdb", times
                    #for t in times: print '%.1f' % t
                    #print "h5parm", sf.time
                    #for t in sf.time: print '%.1f' % t
                    try:
                        data_out[solEntry]['values'] = val.T.reshape(shape)
                    except ValueError as err:
                        logging.critical('Mismatch between parmdb table and H5parm '
                        'solution table: Differing number of frequencies and/or times')
                        sys.exit(1)
                ipbar += 1
                pbar.update(ipbar)
            else:
                # Handle TECScreen parmdb
                #
                # Get timewidths, freqwidth and freq from first (non-TEC, phase)
                # solentry
                for nonTECsolType in pdbSolTypes:
                    if nonTECsolType != 'TECScreen' and 'Phase' in nonTECsolType:
                        break
                parmname = pdb_in.getNames(nonTECsolType+':*')[0]
                timewidths = pdb_in.getValuesGrid(parmname)[parmname]['timewidths']
                freqwidth = pdb.getValuesGrid(parmname)[parmname]['freqwidths'][0]
                freq = pdb.getValuesGrid(parmname)[parmname]['freqs'][0]
                data_out = makeTECparmdb(h5parm_in, solset, st_tec, timewidths, freq, freqwidth)

            pdb_out.addValues(data_out)

        pbar.finish()

    logging.info('Done.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading TECSINGLEFREQ module.')

def _run_parser(soltab, parser, step):
    soltabOut = parser.getstr( step, 'soltabOut', 'tec000' )
    refAnt = parser.getstr( step, 'refAnt', '')

    parser.checkSpelling( step, soltab, ['soltabOut', 'refAnt'])
    return run(soltab, soltabOut, refAnt)


def run( soltab, soltabOut='tec000', refAnt='' ):
    """
    Estimate TEC from ph solutions assuming no wrap for solution at t=0

    Parameters
    ----------
    soltabOut : str, optional
        output table name (same solset), by deault "tec".

    refAnt : str, optional
        Reference antenna, by default the first.

    """

    import numpy as np
    import scipy.optimize
    from losoto.lib_unwrap import unwrap

    logging.info("Find TEC for soltab: "+soltab.name)

    # input check
    solType = soltab.getType()
    if solType != 'phase':
       logging.warning("Soltab type of "+soltab._v_name+" is of type "+solType+", should be phase. Ignoring.")
       return 1

    ants = soltab.getAxisValues('ant')
    if refAnt != '' and refAnt != 'closest' and not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+ants[1])
        refAnt = ants[0]
    if refAnt == '': refAnt = ants[0]

    # times and ants needs to be complete or selection is much slower
    times = soltab.getAxisValues('time')

    # create new table
    solset = soltab.getSolset()
    soltabout = solset.makeSoltab(soltype = 'tec', soltabName = soltabOut, axesNames=['ant','time','dir'], \
                      axesVals=[soltab.getAxisValues(axisName) for axisName in ['ant','time','dir']], \
                      vals=np.zeros(shape=(soltab.getAxisLen('ant'),soltab.getAxisLen('time'),soltab.getAxisLen('dir'))), \
                      weights=np.ones(shape=(soltab.getAxisLen('ant'),soltab.getAxisLen('time'),soltab.getAxisLen('dir'))) )
    soltabout.addHistory('Created by TEC operation from %s.' % soltab.name)
        
    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=['freq','time'], weight=True, refAnt=refAnt):

        assert len(coord['freq']) == 1 # it works with phase at only 1 freq

        if not coord['ant'] == refAnt:

            if (weights == 0.).all() == True:
                logging.warning('Skipping flagged antenna: '+coord['ant'])
                fitweights[:] = 0
            else:

                # unwrap
                vals = np.reshape(unwrap(np.squeeze(vals)), vals.shape)
                vals *= coord['freq']/(-8.44797245e9)
                logging.info('%s: average tec: %f TECU' % (coord['ant'], np.mean(vals)))

        # reorder axes back to the original order, needed for setValues
        soltabout.setSelection(ant=coord['ant'], dir=coord['dir'])
        soltabout.setValues( vals )
        soltabout.setValues( weights, weight=True )

    return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This tool is used to split an H5parm into multiple smaller ones.

import sys, os
import numpy as np
from losoto.h5parm import h5parm
from losoto import _version, _logging

_author = "David Rafferty (drafferty@hs.uni-hamburg.de)"

# get input
import argparse
parser = argparse.ArgumentParser(description='Split h5parm files - '+_author)#, version=_version.__version__
parser.add_argument('h5parmFile', help='Input h5parm')
parser.add_argument('axis', help='Axis name to split')
parser.add_argument('length', help='Target length of axis in each output h5parm (in the same units as the axis)')
parser.add_argument('--insolset', '-s', default='sol000', dest='insolset', help='Input solset name [default: sol000]')
parser.add_argument('--outsolset', '-o', default=None, dest='outsolset', help='Output solset name [default: same as insolset]')
parser.add_argument('--root', '-r', default='output', dest='root', help='Root of output h5parm names; e.g., root_0.h5, root_1.h5, etc. [default: output]')
parser.add_argument('--verbose', '-V', '-v', default=False, action='store_true', help='Go Vebose! (default=False)')
parser.add_argument('--clobber', '-c', default=False, action='store_true', help='Replace exising outh5parm files (default=False)')
args = parser.parse_args()

if len(args.h5parmFile) < 1:
    parser.print_help()
    sys.exit(1)

# log
logger = _logging.Logger('info')
logging = _logging.logger
if args.verbose: logger.set_level("debug")

################################

# read input table
h5 = h5parm(args.h5parmFile, readonly=True)
solset = h5.getSolset(args.insolset)
insoltabs = solset.getSoltabNames()
soltabs = []
for insoltab in insoltabs:
    soltab = solset.getSoltab(insoltab)
    soltabs.append( soltab )
pointingNames = []; antennaNames = []
pointingDirections = []; antennaPositions = []
ants = solset.getAnt()
sous = solset.getSou()
for k,v in list(sous.items()):
    if k not in pointingNames:
        pointingNames.append(k)
        pointingDirections.append(v)
for k, v in list(ants.items()):
    if k not in antennaNames:
        antennaNames.append(k)
        antennaPositions.append(v)

# check that all soltabs have the requested axis and decide how many output files to make
for s in soltabs:
    if args.axis not in s.getAxesNames():
        logging.critical("Axis '{0}' not present in soltab '{1}'.".format(args.axis, s.name))
        sys.exit(1)

    vals = s.getAxisValues(args.axis)

    # delineate the chunks
    chunk_ind = np.where(((vals-vals[0]) % float(args.length))[1:] < np.diff(vals))[0]
    nchunks = len(chunk_ind) + 1

# fill the output files
for i in range(nchunks):
    outh5file = '{0}_{1}.h5'.format(args.root, i)
    if os.path.exists(outh5file) and args.clobber:
        os.remove(outh5file)
    logging.info("\nCreating h5parm {}".format(outh5file))
    outh5 = h5parm(outh5file, readonly=False)
    solsetOutName = args.outsolset
    if solsetOutName is None:
        solsetOutName = args.insolset
    solsetOut = outh5.makeSolset(solsetOutName)

    for s in soltabs:
        vals = s.getAxisValues(args.axis)
        if i == 0:
            startval = vals[0]
        else:
            startval = vals[chunk_ind[i-1]+1]
        if i == nchunks-1:
            endval = vals[-1]
        else:
            endval = vals[chunk_ind[i]+1]
        s.setSelection(**{args.axis:{'min':startval,'max':endval,'step':1}})
        solsetOut.makeSoltab(s.getType(), s.name, axesNames=s.getAxesNames(),
                         axesVals=[s.getAxisValues(a) for a in s.getAxesNames()],
                         vals=s.getValues(retAxesVals=False),
                         weights=s.getValues(weight=True, retAxesVals=False))
        s.clearSelection()

    sourceTable = solsetOut.obj._f_get_child('source')
    antennaTable = solsetOut.obj._f_get_child('antenna')
    antennaTable.append(list(zip(*(antennaNames,antennaPositions))))
    sourceTable.append(list(zip(*(pointingNames,pointingDirections))))
    outh5.close()
